# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import unittest
from typing import Iterable, Union

from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.serializers import neo4_serializer


class TestSerialize(unittest.TestCase):

    def test_serialize(self) -> None:
        actors = [Actor('Tom Cruise'), Actor('Meg Ryan')]
        cities = [City('San Diego'), City('Oakland')]
        movie = Movie('Top Gun', actors, cities)

        actual = []
        node_row = movie.next_node()
        while node_row:
            actual.append(neo4_serializer.serialize_node(node_row))
            node_row = movie.next_node()

        expected = [
            {'name': 'Top Gun', 'KEY': 'movie://Top Gun', 'LABEL': 'Movie'},
            {'name': 'Top Gun', 'KEY': 'actor://Tom Cruise', 'LABEL': 'Actor'},
            {'name': 'Top Gun', 'KEY': 'actor://Meg Ryan', 'LABEL': 'Actor'},
            {'name': 'Top Gun', 'KEY': 'city://San Diego', 'LABEL': 'City'},
            {'name': 'Top Gun', 'KEY': 'city://Oakland', 'LABEL': 'City'}
        ]
        self.assertEqual(expected, actual)

        actual = []
        relation_row = movie.next_relation()
        while relation_row:
            actual.append(neo4_serializer.serialize_relationship(relation_row))
            relation_row = movie.next_relation()

        expected = [
            {'END_KEY': 'actor://Tom Cruise', 'START_LABEL': 'Movie',
             'END_LABEL': 'Actor', 'START_KEY': 'movie://Top Gun',
             'TYPE': 'ACTOR', 'REVERSE_TYPE': 'ACTED_IN'},
            {'END_KEY': 'actor://Meg Ryan', 'START_LABEL': 'Movie',
             'END_LABEL': 'Actor', 'START_KEY': 'movie://Top Gun',
             'TYPE': 'ACTOR', 'REVERSE_TYPE': 'ACTED_IN'},
            {'END_KEY': 'city://San Diego', 'START_LABEL': 'Movie',
             'END_LABEL': 'City', 'START_KEY': 'movie://Top Gun',
             'TYPE': 'FILMED_AT', 'REVERSE_TYPE': 'APPEARS_IN'},
            {'END_KEY': 'city://Oakland', 'START_LABEL': 'Movie',
             'END_LABEL': 'City', 'START_KEY': 'movie://Top Gun',
             'TYPE': 'FILMED_AT', 'REVERSE_TYPE': 'APPEARS_IN'}
        ]
        self.assertEqual(expected, actual)


class Actor(object):
    LABEL = 'Actor'
    KEY_FORMAT = 'actor://{}'

    def __init__(self, name: str) -> None:
        self.name = name


class City(object):
    LABEL = 'City'
    KEY_FORMAT = 'city://{}'

    def __init__(self, name: str) -> None:
        self.name = name


class Movie(GraphSerializable):
    LABEL = 'Movie'
    KEY_FORMAT = 'movie://{}'
    MOVIE_ACTOR_RELATION_TYPE = 'ACTOR'
    ACTOR_MOVIE_RELATION_TYPE = 'ACTED_IN'
    MOVIE_CITY_RELATION_TYPE = 'FILMED_AT'
    CITY_MOVIE_RELATION_TYPE = 'APPEARS_IN'

    def __init__(self,
                 name: str,
                 actors: Iterable[Actor],
                 cities: Iterable[City]) -> None:
        self._name = name
        self._actors = actors
        self._cities = cities
        self._node_iter = iter(self.create_nodes())
        self._relation_iter = iter(self.create_relation())

    def create_next_node(self) -> Union[GraphNode, None]:
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self) -> Union[GraphRelationship, None]:
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    def create_nodes(self) -> Iterable[GraphNode]:
        result = [GraphNode(
            key=Movie.KEY_FORMAT.format(self._name),
            label=Movie.LABEL,
            attributes={
                'name': self._name
            }
        )]

        for actor in self._actors:
            actor_node = GraphNode(
                key=Actor.KEY_FORMAT.format(actor.name),
                label=Actor.LABEL,
                attributes={
                    'name': self._name
                }
            )
            result.append(actor_node)

        for city in self._cities:
            city_node = GraphNode(
                key=City.KEY_FORMAT.format(city.name),
                label=City.LABEL,
                attributes={
                    'name': self._name
                }
            )
            result.append(city_node)
        return result

    def create_relation(self) -> Iterable[GraphRelationship]:
        result = []
        for actor in self._actors:
            movie_actor_relation = GraphRelationship(
                start_key=Movie.KEY_FORMAT.format(self._name),
                end_key=Actor.KEY_FORMAT.format(actor.name),
                start_label=Movie.LABEL,
                end_label=Actor.LABEL,
                type=Movie.MOVIE_ACTOR_RELATION_TYPE,
                reverse_type=Movie.ACTOR_MOVIE_RELATION_TYPE,
                attributes={}
            )
            result.append(movie_actor_relation)

        for city in self._cities:
            city_movie_relation = GraphRelationship(
                start_key=Movie.KEY_FORMAT.format(self._name),
                end_key=City.KEY_FORMAT.format(city.name),
                start_label=Movie.LABEL,
                end_label=City.LABEL,
                type=Movie.MOVIE_CITY_RELATION_TYPE,
                reverse_type=Movie.CITY_MOVIE_RELATION_TYPE,
                attributes={}
            )
            result.append(city_movie_relation)
        return result


if __name__ == '__main__':
    unittest.main()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import unittest
from typing import (
    Any, Iterable, List, Optional,
)

from pyhocon import ConfigFactory, ConfigTree

from databuilder.extractor.base_extractor import Extractor
from databuilder.job.job import DefaultJob
from databuilder.loader.base_loader import Loader
from databuilder.models.table_metadata import TableMetadata
from databuilder.models.table_owner import TableOwner
from databuilder.task.task import DefaultTask
from databuilder.transformer.base_transformer import (
    ChainedTransformer, NoopTransformer, Transformer,
)

TEST_DATA = [
    TableMetadata(
        database="db1", schema="schema1", name="table1", cluster="prod", description=""
    ),
    TableMetadata(
        database="db2", schema="schema2", name="table2", cluster="prod", description=""
    ),
]

EXPECTED_OWNERS = [
    TableOwner(
        db_name="db1",
        cluster="prod",
        schema="schema1",
        table_name="table1",
        owners=["foo", "bar"],
    ),
    TableOwner(
        db_name="db2",
        cluster="prod",
        schema="schema2",
        table_name="table2",
        owners=["foo", "bar"],
    ),
]


class TestChainedTransformerTask(unittest.TestCase):
    def test_multi_yield_task(self) -> None:
        """ Test that MultiYieldTask is able to unpack a transformer which yields multiple nodes """

        result = _run_transformer(AddFakeOwnerTransformer())

        expected = [TEST_DATA[0], EXPECTED_OWNERS[0], TEST_DATA[1], EXPECTED_OWNERS[1]]

        self.assertEqual(repr(result), repr(expected))

    def test_multi_yield_chained_transformer(self) -> None:
        """
        Test that MultiYieldChainedTransformer is able handle both:
            - transformers which yield multiple nodes
            - transformers which transform single nodes
        """

        transformer = ChainedTransformer(
            [AddFakeOwnerTransformer(), NoopTransformer(), DuplicateTransformer()]
        )

        result = _run_transformer(transformer)

        expected = [
            TEST_DATA[0],
            TEST_DATA[0],
            EXPECTED_OWNERS[0],
            EXPECTED_OWNERS[0],
            TEST_DATA[1],
            TEST_DATA[1],
            EXPECTED_OWNERS[1],
            EXPECTED_OWNERS[1],
        ]

        self.assertEqual(repr(result), repr(expected))


class AddFakeOwnerTransformer(Transformer):
    """ A transformer which yields the input record, and also a TableOwner """

    def init(self, conf: ConfigTree) -> None:
        pass

    def get_scope(self) -> str:
        return "transformer.fake_owner"

    def transform(self, record: Any) -> Iterable[Any]:
        yield record
        if isinstance(record, TableMetadata):
            yield TableOwner(
                db_name=record.database,
                schema=record.schema,
                table_name=record.name,
                cluster=record.cluster,
                owners=["foo", "bar"],
            )


class DuplicateTransformer(Transformer):
    """ A transformer which yields the input record twice"""

    def init(self, conf: ConfigTree) -> None:
        pass

    def get_scope(self) -> str:
        return "transformer.duplicate"

    def transform(self, record: Any) -> Iterable[Any]:
        yield record
        yield record


class ListExtractor(Extractor):
    """ An extractor which yields a list of records """

    def init(self, conf: ConfigTree) -> None:
        self.items = conf.get("items")

    def extract(self) -> Optional[Any]:
        try:
            return self.items.pop(0)
        except IndexError:
            return None

    def get_scope(self) -> str:
        return "extractor.test"


class ListLoader(Loader):
    """ A loader which appends all records to a list """

    def init(self, conf: ConfigTree) -> None:
        self.loaded: List[Any] = []

    def load(self, record: Any) -> None:
        self.loaded.append(record)


def _run_transformer(transformer: Transformer) -> List[Any]:
    job_config = ConfigFactory.from_dict({"extractor.test.items": TEST_DATA})

    loader = ListLoader()
    task = DefaultTask(
        extractor=ListExtractor(), transformer=transformer, loader=loader
    )
    job = DefaultJob(conf=job_config, task=task)

    job.launch()
    return loader.loaded



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import unittest
from unittest.mock import ANY

from databuilder.models.schema.schema import SchemaModel
from databuilder.serializers import (
    mysql_serializer, neo4_serializer, neptune_serializer,
)
from databuilder.serializers.neptune_serializer import (
    METADATA_KEY_PROPERTY_NAME, NEPTUNE_CREATION_TYPE_JOB, NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME_BULK_LOADER_FORMAT,
    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT, NEPTUNE_HEADER_ID, NEPTUNE_HEADER_LABEL,
    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT, NEPTUNE_RELATIONSHIP_HEADER_FROM,
    NEPTUNE_RELATIONSHIP_HEADER_TO,
)


class TestSchemaDescription(unittest.TestCase):
    def setUp(self) -> None:
        self.schema = SchemaModel(
            schema_key='db://cluster.schema',
            schema='schema_name',
            description='foo'
        )

    def test_create_nodes(self) -> None:
        schema_node = self.schema.create_next_node()
        serialized_schema_node = neo4_serializer.serialize_node(schema_node)
        schema_desc_node = self.schema.create_next_node()
        serialized_schema_desc_node = neo4_serializer.serialize_node(schema_desc_node)
        self.assertDictEqual(
            serialized_schema_node,
            {'name': 'schema_name', 'KEY': 'db://cluster.schema', 'LABEL': 'Schema'}
        )
        self.assertDictEqual(serialized_schema_desc_node,
                             {'description_source': 'description', 'description': 'foo',
                              'KEY': 'db://cluster.schema/_description', 'LABEL': 'Description'}
                             )
        self.assertIsNone(self.schema.create_next_node())

    def test_create_nodes_neptune(self) -> None:
        schema_node = self.schema.create_next_node()
        expected_serialized_schema_node = {
            NEPTUNE_HEADER_ID: 'Schema:db://cluster.schema',
            METADATA_KEY_PROPERTY_NAME: 'Schema:db://cluster.schema',
            NEPTUNE_HEADER_LABEL: 'Schema',
            NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
            NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB,
            'name:String(single)': 'schema_name',
        }
        serialized_schema_node = neptune_serializer.convert_node(schema_node)
        self.assertDictEqual(
            expected_serialized_schema_node,
            serialized_schema_node
        )
        schema_desc_node = self.schema.create_next_node()
        excepted_serialized_schema_desc_node = {
            NEPTUNE_HEADER_ID: 'Description:db://cluster.schema/_description',
            METADATA_KEY_PROPERTY_NAME: 'Description:db://cluster.schema/_description',
            NEPTUNE_HEADER_LABEL: 'Description',
            NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
            NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB,
            'description_source:String(single)': 'description',
            'description:String(single)': 'foo',
        }
        serialized_schema_desc_node = neptune_serializer.convert_node(schema_desc_node)
        self.assertDictEqual(
            excepted_serialized_schema_desc_node,
            serialized_schema_desc_node
        )

    def test_create_nodes_no_description(self) -> None:

        schema = SchemaModel(schema_key='db://cluster.schema',
                             schema='schema_name')

        schema_node = schema.create_next_node()
        serialized_schema_node = neo4_serializer.serialize_node(schema_node)

        self.assertDictEqual(serialized_schema_node,
                             {'name': 'schema_name', 'KEY': 'db://cluster.schema', 'LABEL': 'Schema'})
        self.assertIsNone(schema.create_next_node())

    def test_create_nodes_programmatic_description(self) -> None:

        schema = SchemaModel(schema_key='db://cluster.schema',
                             schema='schema_name',
                             description='foo',
                             description_source='bar')

        schema_node = schema.create_next_node()
        serialized_schema_node = neo4_serializer.serialize_node(schema_node)
        schema_desc_node = schema.create_next_node()
        serialized_schema_prod_desc_node = neo4_serializer.serialize_node(schema_desc_node)

        self.assertDictEqual(serialized_schema_node,
                             {'name': 'schema_name', 'KEY': 'db://cluster.schema', 'LABEL': 'Schema'})
        self.assertDictEqual(serialized_schema_prod_desc_node,
                             {'description_source': 'bar', 'description': 'foo',
                              'KEY': 'db://cluster.schema/_bar_description', 'LABEL': 'Programmatic_Description'})
        self.assertIsNone(schema.create_next_node())

    def test_create_relation(self) -> None:
        actual = self.schema.create_next_relation()
        serialized_actual = neo4_serializer.serialize_relationship(actual)
        expected = {'END_KEY': 'db://cluster.schema/_description', 'START_LABEL': 'Schema', 'END_LABEL': 'Description',
                    'START_KEY': 'db://cluster.schema', 'TYPE': 'DESCRIPTION', 'REVERSE_TYPE': 'DESCRIPTION_OF'}

        self.assertEqual(expected, serialized_actual)
        self.assertIsNone(self.schema.create_next_relation())

    def test_create_relation_neptune(self) -> None:
        actual = self.schema.create_next_relation()
        serialized_actual = neptune_serializer.convert_relationship(actual)
        forward_header_id = "{label}:{from_vertex_id}_{to_vertex_id}".format(
            from_vertex_id='Schema:db://cluster.schema',
            to_vertex_id='Description:db://cluster.schema/_description',
            label='DESCRIPTION'
        )
        reverse_header_id = "{label}:{from_vertex_id}_{to_vertex_id}".format(
            from_vertex_id='Description:db://cluster.schema/_description',
            to_vertex_id='Schema:db://cluster.schema',
            label='DESCRIPTION_OF'
        )

        neptune_forward_expected = {
            NEPTUNE_HEADER_ID: forward_header_id,
            METADATA_KEY_PROPERTY_NAME: forward_header_id,
            NEPTUNE_RELATIONSHIP_HEADER_FROM: 'Schema:db://cluster.schema',
            NEPTUNE_RELATIONSHIP_HEADER_TO: 'Description:db://cluster.schema/_description',
            NEPTUNE_HEADER_LABEL: 'DESCRIPTION',
            NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
            NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
        }

        neptune_reversed_expected = {
            NEPTUNE_HEADER_ID: reverse_header_id,
            METADATA_KEY_PROPERTY_NAME: reverse_header_id,
            NEPTUNE_RELATIONSHIP_HEADER_FROM: 'Description:db://cluster.schema/_description',
            NEPTUNE_RELATIONSHIP_HEADER_TO: 'Schema:db://cluster.schema',
            NEPTUNE_HEADER_LABEL: 'DESCRIPTION_OF',
            NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
            NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
        }

        self.maxDiff = None
        self.assertDictEqual(serialized_actual[0], neptune_forward_expected)
        self.assertDictEqual(serialized_actual[1], neptune_reversed_expected)

    def test_create_relation_no_description(self) -> None:
        schema = SchemaModel(schema_key='db://cluster.schema',
                             schema='schema_name')

        self.assertIsNone(schema.create_next_relation())

    def test_create_relation_programmatic_description(self) -> None:
        schema = SchemaModel(schema_key='db://cluster.schema',
                             schema='schema_name',
                             description='foo',
                             description_source='bar')

        actual = schema.create_next_relation()
        serialized_actual = neo4_serializer.serialize_relationship(actual)
        expected = {
            'END_KEY': 'db://cluster.schema/_bar_description', 'START_LABEL': 'Schema',
            'END_LABEL': 'Programmatic_Description', 'START_KEY': 'db://cluster.schema', 'TYPE': 'DESCRIPTION',
            'REVERSE_TYPE': 'DESCRIPTION_OF'
        }

        self.assertEqual(expected, serialized_actual)
        self.assertIsNone(schema.create_next_relation())

    def test_create_records(self) -> None:
        schema_record = self.schema.create_next_record()
        serialized_schema_record = mysql_serializer.serialize_record(schema_record)
        schema_desc_record = self.schema.create_next_record()
        serialized_schema_desc_record = mysql_serializer.serialize_record(schema_desc_record)
        self.assertDictEqual(serialized_schema_record, {'rk': 'db://cluster.schema', 'name': 'schema_name',
                                                        'cluster_rk': 'db://cluster'})
        self.assertDictEqual(serialized_schema_desc_record, {'rk': 'db://cluster.schema/_description',
                                                             'description_source': 'description', 'description': 'foo',
                                                             'schema_rk': 'db://cluster.schema'})

    def test_create_records_no_description(self) -> None:
        schema = SchemaModel(schema_key='db://cluster.schema',
                             schema='schema_name')
        schema_record = schema.create_next_record()
        serialized_schema_record = mysql_serializer.serialize_record(schema_record)
        self.assertDictEqual(serialized_schema_record, {'rk': 'db://cluster.schema', 'name': 'schema_name',
                                                        'cluster_rk': 'db://cluster'})
        self.assertIsNone(schema.create_next_record())

    def test_create_records_programmatic_description(self) -> None:
        schema = SchemaModel(schema_key='db://cluster.schema',
                             schema='schema_name',
                             description='foo',
                             description_source='bar')

        schema_record = schema.create_next_record()
        serialized_schema_record = mysql_serializer.serialize_record(schema_record)
        schema_prog_desc_record = schema.create_next_record()
        serialized_schema_prog_desc_record = mysql_serializer.serialize_record(schema_prog_desc_record)
        self.assertDictEqual(serialized_schema_record, {'rk': 'db://cluster.schema', 'name': 'schema_name',
                                                        'cluster_rk': 'db://cluster'})
        self.assertDictEqual(serialized_schema_prog_desc_record, {'rk': 'db://cluster.schema/_bar_description',
                                                                  'description_source': 'bar',
                                                                  'description': 'foo',
                                                                  'schema_rk': 'db://cluster.schema'})



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import unittest

from pyhocon import ConfigFactory

from databuilder import Scoped
from databuilder.extractor.generic_extractor import GenericExtractor


class TestGenericExtractor(unittest.TestCase):

    def test_extraction_with_model_class(self) -> None:
        """
        Test Extraction using model class
        """
        config_dict = {
            'extractor.generic.extraction_items': [{'timestamp': 10000000}],
            'extractor.generic.model_class': 'databuilder.models.es_last_updated.ESLastUpdated',
        }
        conf = ConfigFactory.from_dict(config_dict)

        extractor = GenericExtractor()
        self.conf = ConfigFactory.from_dict(config_dict)
        extractor.init(Scoped.get_scoped_conf(conf=conf,
                                              scope=extractor.get_scope()))

        result = extractor.extract()
        self.assertEqual(result.timestamp, 10000000)

    def test_extraction_without_model_class(self) -> None:
        """
        Test Extraction using model class
        """
        config_dict = {
            'extractor.generic.extraction_items': [{'foo': 1}, {'bar': 2}],
        }
        conf = ConfigFactory.from_dict(config_dict)

        extractor = GenericExtractor()
        self.conf = ConfigFactory.from_dict(config_dict)
        extractor.init(Scoped.get_scoped_conf(conf=conf,
                                              scope=extractor.get_scope()))

        self.assertEqual(extractor.extract(), {'foo': 1})
        self.assertEqual(extractor.extract(), {'bar': 2})



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import json
import unittest

from databuilder.models.dashboard_elasticsearch_document import DashboardESDocument


class TestDashboardElasticsearchDocument(unittest.TestCase):

    def test_to_json(self) -> None:
        """
        Test string generated from to_json method
        """
        test_obj = DashboardESDocument(group_name='test_dashboard_group',
                                       name='test_dashboard_name',
                                       description='test_description',
                                       product='mode',
                                       cluster='gold',
                                       group_description='work space group',
                                       query_names=['query1'],
                                       chart_names=['chart1'],
                                       group_url='mode_group_url',
                                       url='mode_report_url',
                                       uri='mode_dashboard://gold.cluster/dashboard_group/dashboard',
                                       last_successful_run_timestamp=10,
                                       total_usage=10,
                                       tags=['test'],
                                       badges=['test_badge'])

        expected_document_dict = {"group_name": "test_dashboard_group",
                                  "name": "test_dashboard_name",
                                  "description": "test_description",
                                  "product": "mode",
                                  "cluster": "gold",
                                  "group_url": "mode_group_url",
                                  "url": "mode_report_url",
                                  "uri": "mode_dashboard://gold.cluster/dashboard_group/dashboard",
                                  "query_names": ['query1'],
                                  "chart_names": ['chart1'],
                                  "last_successful_run_timestamp": 10,
                                  "group_description": "work space group",
                                  "total_usage": 10,
                                  "tags": ["test"],
                                  "badges": ["test_badge"],

                                  }

        result = test_obj.to_json()
        results = result.split("\n")

        # verify two new line characters in result
        self.assertEqual(len(results), 2, "Result from to_json() function doesn't have a newline!")
        self.assertDictEqual(json.loads(results[0]), expected_document_dict)



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
import unittest
from typing import Any, Dict

from mock import MagicMock, patch
from pyhocon import ConfigFactory

from databuilder.extractor.postgres_metadata_extractor import PostgresMetadataExtractor
from databuilder.extractor.sql_alchemy_extractor import SQLAlchemyExtractor
from databuilder.models.table_metadata import ColumnMetadata, TableMetadata


class TestPostgresMetadataExtractor(unittest.TestCase):
    def setUp(self) -> None:
        logging.basicConfig(level=logging.INFO)

        config_dict = {
            f'extractor.sqlalchemy.{SQLAlchemyExtractor.CONN_STRING}': 'TEST_CONNECTION',
            PostgresMetadataExtractor.CLUSTER_KEY: 'MY_CLUSTER',
            PostgresMetadataExtractor.USE_CATALOG_AS_CLUSTER_NAME: False,
            PostgresMetadataExtractor.DATABASE_KEY: 'postgres'
        }
        self.conf = ConfigFactory.from_dict(config_dict)

    def test_extraction_with_empty_query_result(self) -> None:
        """
        Test Extraction with empty result from query
        """
        with patch.object(SQLAlchemyExtractor, '_get_connection'):
            extractor = PostgresMetadataExtractor()
            extractor.init(self.conf)

            results = extractor.extract()
            self.assertEqual(results, None)

    def test_extraction_with_single_result(self) -> None:
        with patch.object(SQLAlchemyExtractor, '_get_connection') as mock_connection:
            connection = MagicMock()
            mock_connection.return_value = connection
            sql_execute = MagicMock()
            connection.execute = sql_execute
            table = {'schema': 'test_schema',
                     'name': 'test_table',
                     'description': 'a table for testing',
                     'cluster':
                         self.conf[PostgresMetadataExtractor.CLUSTER_KEY]
                     }

            sql_execute.return_value = [
                self._union(
                    {'col_name': 'col_id1',
                     'col_type': 'bigint',
                     'col_description': 'description of id1',
                     'col_sort_order': 0}, table),
                self._union(
                    {'col_name': 'col_id2',
                     'col_type': 'bigint',
                     'col_description': 'description of id2',
                     'col_sort_order': 1}, table),
                self._union(
                    {'col_name': 'is_active',
                     'col_type': 'boolean',
                     'col_description': None,
                     'col_sort_order': 2}, table),
                self._union(
                    {'col_name': 'source',
                     'col_type': 'varchar',
                     'col_description': 'description of source',
                     'col_sort_order': 3}, table),
                self._union(
                    {'col_name': 'etl_created_at',
                     'col_type': 'timestamp',
                     'col_description': 'description of etl_created_at',
                     'col_sort_order': 4}, table),
                self._union(
                    {'col_name': 'ds',
                     'col_type': 'varchar',
                     'col_description': None,
                     'col_sort_order': 5}, table)
            ]

            extractor = PostgresMetadataExtractor()
            extractor.init(self.conf)
            actual = extractor.extract()
            expected = TableMetadata('postgres', 'MY_CLUSTER', 'test_schema', 'test_table', 'a table for testing',
                                     [ColumnMetadata('col_id1', 'description of id1', 'bigint', 0),
                                      ColumnMetadata('col_id2', 'description of id2', 'bigint', 1),
                                      ColumnMetadata('is_active', None, 'boolean', 2),
                                      ColumnMetadata('source', 'description of source', 'varchar', 3),
                                      ColumnMetadata('etl_created_at', 'description of etl_created_at', 'timestamp', 4),
                                      ColumnMetadata('ds', None, 'varchar', 5)])

            self.assertEqual(expected.__repr__(), actual.__repr__())
            self.assertIsNone(extractor.extract())

    def test_extraction_with_multiple_result(self) -> None:
        with patch.object(SQLAlchemyExtractor, '_get_connection') as mock_connection:
            connection = MagicMock()
            mock_connection.return_value = connection
            sql_execute = MagicMock()
            connection.execute = sql_execute
            table = {'schema': 'test_schema1',
                     'name': 'test_table1',
                     'description': 'test table 1',
                     'cluster':
                         self.conf[PostgresMetadataExtractor.CLUSTER_KEY]
                     }

            table1 = {'schema': 'test_schema1',
                      'name': 'test_table2',
                      'description': 'test table 2',
                      'cluster':
                          self.conf[PostgresMetadataExtractor.CLUSTER_KEY]
                      }

            table2 = {'schema': 'test_schema2',
                      'name': 'test_table3',
                      'description': 'test table 3',
                      'cluster':
                          self.conf[PostgresMetadataExtractor.CLUSTER_KEY]
                      }

            sql_execute.return_value = [
                self._union(
                    {'col_name': 'col_id1',
                     'col_type': 'bigint',
                     'col_description': 'description of col_id1',
                     'col_sort_order': 0}, table),
                self._union(
                    {'col_name': 'col_id2',
                     'col_type': 'bigint',
                     'col_description': 'description of col_id2',
                     'col_sort_order': 1}, table),
                self._union(
                    {'col_name': 'is_active',
                     'col_type': 'boolean',
                     'col_description': None,
                     'col_sort_order': 2}, table),
                self._union(
                    {'col_name': 'source',
                     'col_type': 'varchar',
                     'col_description': 'description of source',
                     'col_sort_order': 3}, table),
                self._union(
                    {'col_name': 'etl_created_at',
                     'col_type': 'timestamp',
                     'col_description': 'description of etl_created_at',
                     'col_sort_order': 4}, table),
                self._union(
                    {'col_name': 'ds',
                     'col_type': 'varchar',
                     'col_description': None,
                     'col_sort_order': 5}, table),
                self._union(
                    {'col_name': 'col_name',
                     'col_type': 'varchar',
                     'col_description': 'description of col_name',
                     'col_sort_order': 0}, table1),
                self._union(
                    {'col_name': 'col_name2',
                     'col_type': 'varchar',
                     'col_description': 'description of col_name2',
                     'col_sort_order': 1}, table1),
                self._union(
                    {'col_name': 'col_id3',
                     'col_type': 'varchar',
                     'col_description': 'description of col_id3',
                     'col_sort_order': 0}, table2),
                self._union(
                    {'col_name': 'col_name3',
                     'col_type': 'varchar',
                     'col_description': 'description of col_name3',
                     'col_sort_order': 1}, table2)
            ]

            extractor = PostgresMetadataExtractor()
            extractor.init(self.conf)

            expected = TableMetadata('postgres',
                                     self.conf[PostgresMetadataExtractor.CLUSTER_KEY],
                                     'test_schema1', 'test_table1', 'test table 1',
                                     [ColumnMetadata('col_id1', 'description of col_id1', 'bigint', 0),
                                      ColumnMetadata('col_id2', 'description of col_id2', 'bigint', 1),
                                      ColumnMetadata('is_active', None, 'boolean', 2),
                                      ColumnMetadata('source', 'description of source', 'varchar', 3),
                                      ColumnMetadata('etl_created_at', 'description of etl_created_at', 'timestamp', 4),
                                      ColumnMetadata('ds', None, 'varchar', 5)])
            self.assertEqual(expected.__repr__(), extractor.extract().__repr__())

            expected = TableMetadata('postgres',
                                     self.conf[PostgresMetadataExtractor.CLUSTER_KEY],
                                     'test_schema1', 'test_table2', 'test table 2',
                                     [ColumnMetadata('col_name', 'description of col_name', 'varchar', 0),
                                      ColumnMetadata('col_name2', 'description of col_name2', 'varchar', 1)])
            self.assertEqual(expected.__repr__(), extractor.extract().__repr__())

            expected = TableMetadata('postgres',
                                     self.conf[PostgresMetadataExtractor.CLUSTER_KEY],
                                     'test_schema2', 'test_table3', 'test table 3',
                                     [ColumnMetadata('col_id3', 'description of col_id3', 'varchar', 0),
                                      ColumnMetadata('col_name3', 'description of col_name3',
                                                     'varchar', 1)])
            self.assertEqual(expected.__repr__(), extractor.extract().__repr__())

            self.assertIsNone(extractor.extract())
            self.assertIsNone(extractor.extract())

    def _union(self,
               target: Dict[Any, Any],
               extra: Dict[Any, Any]) -> Dict[Any, Any]:
        target.update(extra)
        return target


class TestPostgresMetadataExtractorWithWhereClause(unittest.TestCase):
    def setUp(self) -> None:
        logging.basicConfig(level=logging.INFO)
        self.where_clause_suffix = """
        where table_schema in ('public') and table_name = 'movies'
        """

        config_dict = {
            PostgresMetadataExtractor.WHERE_CLAUSE_SUFFIX_KEY: self.where_clause_suffix,
            f'extractor.sqlalchemy.{SQLAlchemyExtractor.CONN_STRING}': 'TEST_CONNECTION'
        }
        self.conf = ConfigFactory.from_dict(config_dict)

    def test_sql_statement(self) -> None:
        """
        Test Extraction with empty result from query
        """
        with patch.object(SQLAlchemyExtractor, '_get_connection'):
            extractor = PostgresMetadataExtractor()
            extractor.init(self.conf)
            self.assertTrue(self.where_clause_suffix in extractor.sql_stmt)


class TestPostgresMetadataExtractorClusterKeyNoTableCatalog(unittest.TestCase):
    # test when USE_CATALOG_AS_CLUSTER_NAME is false and CLUSTER_KEY is specified
    def setUp(self) -> None:
        logging.basicConfig(level=logging.INFO)
        self.cluster_key = "not_master"

        config_dict = {
            PostgresMetadataExtractor.CLUSTER_KEY: self.cluster_key,
            f'extractor.sqlalchemy.{SQLAlchemyExtractor.CONN_STRING}': 'TEST_CONNECTION',
            PostgresMetadataExtractor.USE_CATALOG_AS_CLUSTER_NAME: False
        }
        self.conf = ConfigFactory.from_dict(config_dict)

    def test_sql_statement(self) -> None:
        """
        Test Extraction with empty result from query
        """
        with patch.object(SQLAlchemyExtractor, '_get_connection'):
            extractor = PostgresMetadataExtractor()
            extractor.init(self.conf)
            self.assertTrue(self.cluster_key in extractor.sql_stmt)


class TestPostgresMetadataExtractorNoClusterKeyNoTableCatalog(unittest.TestCase):
    # test when USE_CATALOG_AS_CLUSTER_NAME is false and CLUSTER_KEY is NOT specified
    def setUp(self) -> None:
        logging.basicConfig(level=logging.INFO)

        config_dict = {
            f'extractor.sqlalchemy.{SQLAlchemyExtractor.CONN_STRING}': 'TEST_CONNECTION',
            PostgresMetadataExtractor.USE_CATALOG_AS_CLUSTER_NAME: False
        }
        self.conf = ConfigFactory.from_dict(config_dict)

    def test_sql_statement(self) -> None:
        """
        Test Extraction with empty result from query
        """
        with patch.object(SQLAlchemyExtractor, '_get_connection'):
            extractor = PostgresMetadataExtractor()
            extractor.init(self.conf)
            self.assertTrue(PostgresMetadataExtractor.DEFAULT_CLUSTER_NAME in extractor.sql_stmt)


class TestPostgresMetadataExtractorTableCatalogEnabled(unittest.TestCase):
    # test when USE_CATALOG_AS_CLUSTER_NAME is true (CLUSTER_KEY should be ignored)
    def setUp(self) -> None:
        logging.basicConfig(level=logging.INFO)
        self.cluster_key = "not_master"

        config_dict = {
            PostgresMetadataExtractor.CLUSTER_KEY: self.cluster_key,
            f'extractor.sqlalchemy.{SQLAlchemyExtractor.CONN_STRING}': 'TEST_CONNECTION',
            PostgresMetadataExtractor.USE_CATALOG_AS_CLUSTER_NAME: True
        }
        self.conf = ConfigFactory.from_dict(config_dict)

    def test_sql_statement(self) -> None:
        """
        Test Extraction with empty result from query
        """
        with patch.object(SQLAlchemyExtractor, '_get_connection'):
            extractor = PostgresMetadataExtractor()
            extractor.init(self.conf)
            self.assertTrue('table_catalog' in extractor.sql_stmt)
            self.assertFalse(self.cluster_key in extractor.sql_stmt)


if __name__ == '__main__':
    unittest.main()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import unittest

from pyhocon import ConfigFactory

from databuilder.transformer.remove_field_transformer import FIELD_NAMES, RemoveFieldTransformer


class TestRemoveFieldTransformer(unittest.TestCase):

    def test_conversion(self) -> None:

        transformer = RemoveFieldTransformer()
        config = ConfigFactory.from_dict({
            FIELD_NAMES: ['foo', 'bar'],
        })
        transformer.init(conf=config)

        actual = transformer.transform({
            'foo': 'foo_val',
            'bar': 'bar_val',
            'baz': 'baz_val',
        })
        expected = {
            'baz': 'baz_val'
        }
        self.assertDictEqual(expected, actual)

    def test_conversion_missing_field(self) -> None:

        transformer = RemoveFieldTransformer()
        config = ConfigFactory.from_dict({
            FIELD_NAMES: ['foo', 'bar'],
        })
        transformer.init(conf=config)

        actual = transformer.transform({
            'foo': 'foo_val',
            'baz': 'baz_val',
            'john': 'doe',
        })
        expected = {
            'baz': 'baz_val',
            'john': 'doe'
        }
        self.assertDictEqual(expected, actual)


if __name__ == '__main__':
    unittest.main()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
import unittest
from typing import Any

from mock import Mock, patch
from pyhocon import ConfigFactory

from databuilder import Scoped
from databuilder.extractor.bigquery_metadata_extractor import BigQueryMetadataExtractor
from databuilder.models.table_metadata import TableMetadata

logging.basicConfig(level=logging.INFO)

NO_DATASETS = {'kind': 'bigquery#datasetList', 'etag': '1B2M2Y8AsgTpgAmY7PhCfg=='}
ONE_DATASET = {
    'kind': 'bigquery#datasetList', 'etag': 'yScH5WIHeNUBF9b/VKybXA==',
    'datasets': [{
        'kind': 'bigquery#dataset',
        'id': 'your-project-here:empty',
        'datasetReference': {
            'datasetId': 'empty',
            'projectId': 'your-project-here'
        },
        'location': 'US'
    }]
}  # noqa
NO_TABLES = {'kind': 'bigquery#tableList', 'etag': '1B2M2Y8AsgTpgAmY7PhCfg==', 'totalItems': 0}
ONE_TABLE = {
    'kind': 'bigquery#tableList', 'etag': 'Iaqrz2TCDIANAOD/Xerkjw==',
    'tables': [{
        'kind': 'bigquery#table',
        'id': 'your-project-here:fdgdfgh.nested_recs',
        'tableReference': {
            'projectId': 'your-project-here',
            'datasetId': 'fdgdfgh',
            'tableId': 'nested_recs'
        },
        'type': 'TABLE',
        'creationTime': '1557578974009'
    }],
    'totalItems': 1
}  # noqa
ONE_VIEW = {
    'kind': 'bigquery#tableList', 'etag': 'Iaqrz2TCDIANAOD/Xerkjw==',
    'tables': [{
        'kind': 'bigquery#table',
        'id': 'your-project-here:fdgdfgh.abab',
        'tableReference': {'projectId': 'your-project-here', 'datasetId': 'fdgdfgh', 'tableId': 'abab'},
        'type': 'VIEW',
        'view': {'useLegacySql': False},
        'creationTime': '1557577874991'
    }],
    'totalItems': 1
}  # noqa
TIME_PARTITIONED = {
    'kind': 'bigquery#tableList', 'etag': 'Iaqrz2TCDIANAOD/Xerkjw==',
    'tables': [{
        'kind': 'bigquery#table',
        'id': 'your-project-here:fdgdfgh.other',
        'tableReference': {'projectId': 'your-project-here', 'datasetId': 'fdgdfgh', 'tableId': 'other'},
        'type': 'TABLE',
        'timePartitioning': {'type': 'DAY', 'requirePartitionFilter': False},
        'creationTime': '1557577779306'
    }],
    'totalItems': 1
}  # noqa
TABLE_DATE_RANGE = {
    'kind': 'bigquery#tableList', 'etag': 'Iaqrz2TCDIANAOD/Xerkjw==',
    'tables': [{
        'kind': 'bigquery#table', 'id': 'your-project-here:fdgdfgh.other_20190101',
        'tableReference': {'projectId': 'your-project-here', 'datasetId': 'fdgdfgh', 'tableId': 'date_range_20190101'},
        'type': 'TABLE',
        'creationTime': '1557577779306'
    }, {
        'kind': 'bigquery#table', 'id': 'your-project-here:fdgdfgh.other_20190102',
        'tableReference': {'projectId': 'your-project-here', 'datasetId': 'fdgdfgh', 'tableId': 'date_range_20190102'},
        'type': 'TABLE',
        'creationTime': '1557577779306'
    }],
    'totalItems': 2
}  # noqa
TABLE_DATA = {
    'kind': 'bigquery#table', 'etag': 'Hzc/56Rp9VR4Y6jhZApD/g==', 'id': 'your-project-here:fdgdfgh.test',
    'selfLink': 'https://www.googleapis.com/bigquery/v2/projects/your-project-here/datasets/fdgdfgh/tables/test',
    'tableReference': {'projectId': 'your-project-here', 'datasetId': 'fdgdfgh', 'tableId': 'test'},
    'schema': {
        'fields': [{'name': 'test', 'type': 'STRING', 'description': 'some_description'},
                   {'name': 'test2', 'type': 'INTEGER'},
                   {'name': 'test3', 'type': 'FLOAT', 'description': 'another description'},
                   {'name': 'test4', 'type': 'BOOLEAN'},
                   {'name': 'test5', 'type': 'DATETIME'}]
    },
    'numBytes': '0',
    'numLongTermBytes': '0',
    'numRows': '0',
    'creationTime': '1557577756303',
    'lastModifiedTime': '1557577756370',
    'type': 'TABLE',
    'location': 'EU'
}  # noqa
NO_SCHEMA = {
    'kind': 'bigquery#table', 'etag': 'Hzc/56Rp9VR4Y6jhZApD/g==', 'id': 'your-project-here:fdgdfgh.no_schema',
    'selfLink': 'https://www.googleapis.com/bigquery/v2/projects/your-project-here/datasets/fdgdfgh/tables/no_schema',
    'tableReference': {'projectId': 'your-project-here', 'datasetId': 'fdgdfgh', 'tableId': 'no_schema'},
    'numBytes': '0',
    'numLongTermBytes': '0',
    'numRows': '0',
    'creationTime': '1557577756303',
    'lastModifiedTime': '1557577756370',
    'type': 'TABLE',
    'location': 'EU'
}  # noqa
NO_COLS = {
    'kind': 'bigquery#table', 'etag': 'Hzc/56Rp9VR4Y6jhZApD/g==', 'id': 'your-project-here:fdgdfgh.no_columns',
    'selfLink': 'https://www.googleapis.com/bigquery/v2/projects/your-project-here/datasets/fdgdfgh/tables/no_columns',
    'tableReference': {'projectId': 'your-project-here', 'datasetId': 'fdgdfgh', 'tableId': 'no_columns'},
    'schema': {},
    'numBytes': '0',
    'numLongTermBytes': '0',
    'numRows': '0',
    'creationTime': '1557577756303',
    'lastModifiedTime': '1557577756370',
    'type': 'TABLE',
    'location': 'EU'
}  # noqa
VIEW_DATA = {
    'kind': 'bigquery#table', 'etag': 'E6+jjbQ/HsegSNpTEgELUA==', 'id': 'gerard-cloud-2:fdgdfgh.abab',
    'selfLink': 'https://www.googleapis.com/bigquery/v2/projects/gerard-cloud-2/datasets/fdgdfgh/tables/abab',
    'tableReference': {'projectId': 'gerard-cloud-2', 'datasetId': 'fdgdfgh', 'tableId': 'abab'},
    'schema': {
        'fields': [
            {'name': 'test', 'type': 'STRING'},
            {'name': 'test2', 'type': 'INTEGER'},
            {'name': 'test3', 'type': 'FLOAT'},
            {'name': 'test4', 'type': 'BOOLEAN'},
            {'name': 'test5', 'type': 'DATETIME'}]
    },
    'numBytes': '0',
    'numLongTermBytes': '0',
    'numRows': '0',
    'creationTime': '1557577874991',
    'lastModifiedTime': '1557577874991',
    'type': 'VIEW',
    'view': {'query': 'SELECT * from `gerard-cloud-2.fdgdfgh.test`', 'useLegacySql': False},
    'location': 'EU'
}  # noqa
NESTED_DATA = {
    'kind': 'bigquery#table', 'etag': 'Hzc/56Rp9VR4Y6jhZApD/g==', 'id': 'your-project-here:fdgdfgh.test',
    'selfLink': 'https://www.googleapis.com/bigquery/v2/projects/your-project-here/datasets/fdgdfgh/tables/test',
    'tableReference': {'projectId': 'your-project-here', 'datasetId': 'fdgdfgh', 'tableId': 'test'},
    'schema': {
        'fields': [{
            'name': 'nested', 'type': 'RECORD',
            'fields': [{
                'name': 'nested2', 'type': 'RECORD',
                'fields': [{'name': 'ahah', 'type': 'STRING'}]
            }]
        }]
    },
    'type': 'TABLE',
    'location': 'EU'
}  # noqa

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError


class MockBigQueryClient():
    def __init__(self,
                 dataset_list_data: Any,
                 table_list_data: Any,
                 table_data: Any
                 ) -> None:
        self.ds_execute = Mock()
        self.ds_execute.execute.return_value = dataset_list_data
        self.ds_list = Mock()
        self.ds_list.list.return_value = self.ds_execute
        self.list_execute = Mock()
        self.list_execute.execute.return_value = table_list_data
        self.get_execute = Mock()
        self.get_execute.execute.return_value = table_data
        self.tables_method = Mock()
        self.tables_method.list.return_value = self.list_execute
        self.tables_method.get.return_value = self.get_execute

    def datasets(self) -> Any:
        return self.ds_list

    def tables(self) -> Any:
        return self.tables_method


# Patch fallback auth method to avoid actually calling google API
@patch('google.auth.default', lambda scopes: ['dummy', 'dummy'])
class TestBigQueryMetadataExtractor(unittest.TestCase):
    def setUp(self) -> None:
        config_dict = {
            f'extractor.bigquery_table_metadata.{BigQueryMetadataExtractor.PROJECT_ID_KEY}': 'your-project-here'
        }
        self.conf = ConfigFactory.from_dict(config_dict)

    @patch('databuilder.extractor.base_bigquery_extractor.build')
    def test_can_handle_datasets(self, mock_build: Any) -> None:
        mock_build.return_value = MockBigQueryClient(NO_DATASETS, None, None)
        extractor = BigQueryMetadataExtractor()
        extractor.init(Scoped.get_scoped_conf(conf=self.conf,
                                              scope=extractor.get_scope()))
        result = extractor.extract()
        self.assertIsNone(result)

    @patch('databuilder.extractor.base_bigquery_extractor.build')
    def test_empty_dataset(self, mock_build: Any) -> None:
        mock_build.return_value = MockBigQueryClient(ONE_DATASET, NO_TABLES, None)
        extractor = BigQueryMetadataExtractor()
        extractor.init(Scoped.get_scoped_conf(conf=self.conf,
                                              scope=extractor.get_scope()))
        result = extractor.extract()
        self.assertIsNone(result)

    @patch('databuilder.extractor.base_bigquery_extractor.build')
    def test_accepts_dataset_filter_by_label(self, mock_build: Any) -> None:
        config_dict = {
            f'extractor.bigquery_table_metadata.{BigQueryMetadataExtractor.PROJECT_ID_KEY}': 'your-project-here',
            f'extractor.bigquery_table_metadata.{BigQueryMetadataExtractor.FILTER_KEY}': 'label.key:value'
        }
        conf = ConfigFactory.from_dict(config_dict)

        mock_build.return_value = MockBigQueryClient(ONE_DATASET, ONE_TABLE, TABLE_DATA)
        extractor = BigQueryMetadataExtractor()
        extractor.init(Scoped.get_scoped_conf(conf=conf,
                                              scope=extractor.get_scope()))
        result = extractor.extract()
        self.assertIsInstance(result, TableMetadata)

    @patch('databuilder.extractor.base_bigquery_extractor.build')
    def test_table_without_schema(self, mock_build: Any) -> None:
        mock_build.return_value = MockBigQueryClient(ONE_DATASET, ONE_TABLE, NO_SCHEMA)
        extractor = BigQueryMetadataExtractor()
        extractor.init(Scoped.get_scoped_conf(conf=self.conf,
                                              scope=extractor.get_scope()))
        result = extractor.extract()

        self.assertEqual(result.database, 'bigquery')
        self.assertEqual(result.cluster, 'your-project-here')
        self.assertEqual(result.schema, 'fdgdfgh')
        self.assertEqual(result.name, 'nested_recs')
        self.assertEqual(result.description.text, '')
        self.assertEqual(result.columns, [])
        self.assertEqual(result.is_view, False)

    @patch('databuilder.extractor.base_bigquery_extractor.build')
    def test_table_without_columns(self, mock_build: Any) -> None:
        mock_build.return_value = MockBigQueryClient(ONE_DATASET, ONE_TABLE, NO_COLS)
        extractor = BigQueryMetadataExtractor()
        extractor.init(Scoped.get_scoped_conf(conf=self.conf,
                                              scope=extractor.get_scope()))
        result = extractor.extract()

        self.assertEqual(result.database, 'bigquery')
        self.assertEqual(result.cluster, 'your-project-here')
        self.assertEqual(result.schema, 'fdgdfgh')
        self.assertEqual(result.name, 'nested_recs')
        self.assertEqual(result.description.text, "")
        self.assertEqual(result.columns, [])
        self.assertEqual(result.is_view, False)

    @patch('databuilder.extractor.base_bigquery_extractor.build')
    def test_view(self, mock_build: Any) -> None:
        mock_build.return_value = MockBigQueryClient(ONE_DATASET, ONE_VIEW, VIEW_DATA)
        extractor = BigQueryMetadataExtractor()
        extractor.init(Scoped.get_scoped_conf(conf=self.conf,
                                              scope=extractor.get_scope()))
        result = extractor.extract()
        self.assertIsInstance(result, TableMetadata)
        self.assertEqual(result.is_view, True)

    @patch('databuilder.extractor.base_bigquery_extractor.build')
    def test_normal_table(self, mock_build: Any) -> None:
        mock_build.return_value = MockBigQueryClient(ONE_DATASET, ONE_TABLE, TABLE_DATA)
        extractor = BigQueryMetadataExtractor()
        extractor.init(Scoped.get_scoped_conf(conf=self.conf,
                                              scope=extractor.get_scope()))
        result = extractor.extract()

        self.assertEqual(result.database, 'bigquery')
        self.assertEqual(result.cluster, 'your-project-here')
        self.assertEqual(result.schema, 'fdgdfgh')
        self.assertEqual(result.name, 'nested_recs')
        self.assertEqual(result.description.text, "")

        first_col = result.columns[0]
        self.assertEqual(first_col.name, 'test')
        self.assertEqual(first_col.type, 'STRING')
        self.assertEqual(first_col.description.text, 'some_description')
        self.assertEqual(result.is_view, False)

    @patch('databuilder.extractor.base_bigquery_extractor.build')
    def test_table_with_nested_records(self, mock_build: Any) -> None:
        mock_build.return_value = MockBigQueryClient(ONE_DATASET, ONE_TABLE, NESTED_DATA)
        extractor = BigQueryMetadataExtractor()
        extractor.init(Scoped.get_scoped_conf(conf=self.conf,
                                              scope=extractor.get_scope()))
        result = extractor.extract()

        first_col = result.columns[0]
        self.assertEqual(first_col.name, 'nested')
        self.assertEqual(first_col.type, 'RECORD')
        second_col = result.columns[1]
        self.assertEqual(second_col.name, 'nested.nested2')
        self.assertEqual(second_col.type, 'RECORD')
        third_col = result.columns[2]
        self.assertEqual(third_col.name, 'nested.nested2.ahah')
        self.assertEqual(third_col.type, 'STRING')

    @patch('databuilder.extractor.base_bigquery_extractor.build')
    def test_keypath_and_pagesize_can_be_set(self, mock_build: Any) -> None:
        config_dict = {
            f'extractor.bigquery_table_metadata.{BigQueryMetadataExtractor.PROJECT_ID_KEY}': 'your-project-here',
            f'extractor.bigquery_table_metadata.{BigQueryMetadataExtractor.PAGE_SIZE_KEY}': 200,
            f'extractor.bigquery_table_metadata.{BigQueryMetadataExtractor.KEY_PATH_KEY}': '/tmp/doesnotexist',
        }
        conf = ConfigFactory.from_dict(config_dict)

        mock_build.return_value = MockBigQueryClient(ONE_DATASET, ONE_TABLE, TABLE_DATA)
        extractor = BigQueryMetadataExtractor()

        with self.assertRaises(FileNotFoundError):
            extractor.init(Scoped.get_scoped_conf(conf=conf,
                                                  scope=extractor.get_scope()))

    @patch('databuilder.extractor.base_bigquery_extractor.build')
    def test_table_part_of_table_date_range(self, mock_build: Any) -> None:
        mock_build.return_value = MockBigQueryClient(ONE_DATASET, TABLE_DATE_RANGE, TABLE_DATA)
        extractor = BigQueryMetadataExtractor()
        extractor.init(Scoped.get_scoped_conf(conf=self.conf,
                                              scope=extractor.get_scope()))

        count = 0
        result = extractor.extract()
        table_name = result.name
        while result:
            count += 1
            result = extractor.extract()

        self.assertEqual(count, 1)
        self.assertEqual(table_name, 'date_range_')



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import unittest

from pyhocon import ConfigFactory

from databuilder import Scoped
from databuilder.extractor.csv_extractor import CsvExtractor, CsvTableColumnExtractor
from databuilder.models.badge import Badge


class TestCsvExtractor(unittest.TestCase):

    def test_extraction_with_model_class(self) -> None:
        """
        Test Extraction using model class
        """
        config_dict = {
            f'extractor.csv.{CsvExtractor.FILE_LOCATION}': 'example/sample_data/sample_table.csv',
            f'extractor.csv.model_class': 'databuilder.models.table_metadata.TableMetadata',
        }
        self.conf = ConfigFactory.from_dict(config_dict)
        extractor = CsvExtractor()
        extractor.init(Scoped.get_scoped_conf(conf=self.conf,
                                              scope=extractor.get_scope()))

        result = extractor.extract()
        self.assertEqual(result.name, 'test_table1')
        self.assertEqual(result.description.text, '1st test table')
        self.assertEqual(result.database, 'hive')
        self.assertEqual(result.cluster, 'gold')
        self.assertEqual(result.schema, 'test_schema')
        self.assertEqual(result.tags, ['tag1', 'tag2'])
        self.assertEqual(result.is_view, 'false')

        result2 = extractor.extract()
        self.assertEqual(result2.name, 'test_table2')
        self.assertEqual(result2.is_view, 'false')

        result3 = extractor.extract()
        self.assertEqual(result3.name, 'test_view1')
        self.assertEqual(result3.is_view, 'true')

    def test_extraction_of_tablecolumn_badges(self) -> None:
        """
        Test Extraction using the combined CsvTableModel model class
        """
        config_dict = {
            f'extractor.csvtablecolumn.{CsvTableColumnExtractor.TABLE_FILE_LOCATION}':
            'example/sample_data/sample_table.csv',
            f'extractor.csvtablecolumn.{CsvTableColumnExtractor.COLUMN_FILE_LOCATION}':
            'example/sample_data/sample_col.csv',
        }
        self.conf = ConfigFactory.from_dict(config_dict)

        extractor = CsvTableColumnExtractor()
        extractor.init(Scoped.get_scoped_conf(conf=self.conf,
                                              scope=extractor.get_scope()))

        result = extractor.extract()
        self.assertEqual(result.name, 'test_table1')
        self.assertEqual(result.columns[0].badges, [Badge('pk', 'column')])
        self.assertEqual(result.columns[1].badges, [Badge('pii', 'column')])



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import textwrap
import unittest
import uuid

from databuilder.publisher.neo4j_preprocessor import DeleteRelationPreprocessor, NoopRelationPreprocessor


class TestNeo4jPreprocessor(unittest.TestCase):

    def testNoopRelationPreprocessor(self) -> None:
        preprocessor = NoopRelationPreprocessor()

        self.assertTrue(not preprocessor.is_perform_preprocess())

    def testDeleteRelationPreprocessor(self) -> None:  # noqa: W293
        preprocessor = DeleteRelationPreprocessor()

        self.assertTrue(preprocessor.is_perform_preprocess())

        preprocessor.filter(start_label='foo_label',
                            end_label='bar_label',
                            start_key='foo_key',
                            end_key='bar_key',
                            relation='foo_relation',
                            reverse_relation='bar_relation')

        self.assertTrue(preprocessor.filter(start_label=str(uuid.uuid4()),
                                            end_label=str(uuid.uuid4()),
                                            start_key=str(uuid.uuid4()),
                                            end_key=str(uuid.uuid4()),
                                            relation=str(uuid.uuid4()),
                                            reverse_relation=str(uuid.uuid4())))

        actual = preprocessor.preprocess_cypher(start_label='foo_label',
                                                end_label='bar_label',
                                                start_key='foo_key',
                                                end_key='bar_key',
                                                relation='foo_relation',
                                                reverse_relation='bar_relation')

        expected = (textwrap.dedent("""
    MATCH (n1:foo_label {key: $start_key })-[r]-(n2:bar_label {key: $end_key })

    WITH r LIMIT 2
    DELETE r
    RETURN count(*) as count;
    """), {'start_key': 'foo_key', 'end_key': 'bar_key'})

        self.assertEqual(expected, actual)

    def testDeleteRelationPreprocessorFilter(self) -> None:
        preprocessor = DeleteRelationPreprocessor(label_tuples=[('foo', 'bar')])

        self.assertTrue(preprocessor.filter(start_label='foo',
                                            end_label='bar',
                                            start_key=str(uuid.uuid4()),
                                            end_key=str(uuid.uuid4()),
                                            relation=str(uuid.uuid4()),
                                            reverse_relation=str(uuid.uuid4())))

        self.assertTrue(preprocessor.filter(start_label='bar',
                                            end_label='foo',
                                            start_key=str(uuid.uuid4()),
                                            end_key=str(uuid.uuid4()),
                                            relation=str(uuid.uuid4()),
                                            reverse_relation=str(uuid.uuid4())))

        self.assertFalse(preprocessor.filter(start_label='foz',
                                             end_label='baz',
                                             start_key=str(uuid.uuid4()),
                                             end_key=str(uuid.uuid4()),
                                             relation=str(uuid.uuid4()),
                                             reverse_relation=str(uuid.uuid4())))


if __name__ == '__main__':
    unittest.main()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import unittest

from databuilder.models.table_metadata import DescriptionMetadata
from databuilder.serializers import neo4_serializer


class TestDescriptionMetadata(unittest.TestCase):
    def test_raise_exception_when_missing_data(self) -> None:
        # assert raise when missing description node key
        self.assertRaises(
            Exception,
            DescriptionMetadata(text='test_text').next_node
        )
        DescriptionMetadata(text='test_text', description_key='test_key').next_node()
        DescriptionMetadata(text='test_text', start_key='start_key').next_node()

        # assert raise when missing relation start label
        self.assertRaises(
            Exception,
            DescriptionMetadata(text='test_text', start_key='start_key').next_relation
        )
        DescriptionMetadata(text='test_text', start_key='test_key', start_label='Table').next_relation()

        # assert raise when missing relation start key
        self.assertRaises(
            Exception,
            DescriptionMetadata(text='test_text', description_key='test_key', start_label='Table').next_relation
        )

    def test_serialize_table_description_metadata(self) -> None:
        description_metadata = DescriptionMetadata(
            text='test text 1',
            start_label='Table',
            start_key='test_start_key'
        )
        node_row = description_metadata.next_node()
        actual = []
        while node_row:
            node_row_serialized = neo4_serializer.serialize_node(node_row)
            actual.append(node_row_serialized)
            node_row = description_metadata.next_node()
        expected = [
            {'description': 'test text 1', 'KEY': 'test_start_key/_description',
             'LABEL': 'Description', 'description_source': 'description'},
        ]
        self.assertEqual(actual, expected)

        relation_row = description_metadata.next_relation()
        actual = []
        while relation_row:
            relation_row_serialized = neo4_serializer.serialize_relationship(relation_row)
            actual.append(relation_row_serialized)
            relation_row = description_metadata.next_relation()
        expected = [
            {'START_KEY': 'test_start_key', 'START_LABEL': 'Table', 'END_KEY': 'test_start_key/_description',
             'END_LABEL': 'Description', 'TYPE': 'DESCRIPTION', 'REVERSE_TYPE': 'DESCRIPTION_OF'}
        ]
        self.assertEqual(actual, expected)

    def test_serialize_column_description_metadata(self) -> None:
        description_metadata = DescriptionMetadata(
            text='test text 2',
            start_label='Column',
            start_key='test_start_key',
            description_key='customized_key'
        )
        node_row = description_metadata.next_node()
        actual = []
        while node_row:
            node_row_serialized = neo4_serializer.serialize_node(node_row)
            actual.append(node_row_serialized)
            node_row = description_metadata.next_node()
        expected = [
            {'description': 'test text 2', 'KEY': 'customized_key',
             'LABEL': 'Description', 'description_source': 'description'},
        ]
        self.assertEqual(actual, expected)

        relation_row = description_metadata.next_relation()
        actual = []
        while relation_row:
            relation_row_serialized = neo4_serializer.serialize_relationship(relation_row)
            actual.append(relation_row_serialized)
            relation_row = description_metadata.next_relation()
        expected = [
            {'START_KEY': 'test_start_key', 'START_LABEL': 'Column', 'END_KEY': 'customized_key',
             'END_LABEL': 'Description', 'TYPE': 'DESCRIPTION', 'REVERSE_TYPE': 'DESCRIPTION_OF'}
        ]
        self.assertEqual(actual, expected)

    def test_serialize_column_with_source_description_metadata(self) -> None:
        description_metadata = DescriptionMetadata(
            text='test text 3',
            start_label='Column',
            start_key='test_start_key',
            description_key='customized_key',
            source='external',
        )
        node_row = description_metadata.next_node()
        actual = []
        while node_row:
            node_row_serialized = neo4_serializer.serialize_node(node_row)
            actual.append(node_row_serialized)
            node_row = description_metadata.next_node()
        expected = [
            {'description': 'test text 3', 'KEY': 'customized_key',
             'LABEL': 'Programmatic_Description', 'description_source': 'external'},
        ]
        self.assertEqual(actual, expected)

        relation_row = description_metadata.next_relation()
        actual = []
        while relation_row:
            relation_row_serialized = neo4_serializer.serialize_relationship(relation_row)
            actual.append(relation_row_serialized)
            relation_row = description_metadata.next_relation()
        expected = [
            {'START_KEY': 'test_start_key', 'START_LABEL': 'Column', 'END_KEY': 'customized_key',
             'END_LABEL': 'Programmatic_Description', 'TYPE': 'DESCRIPTION', 'REVERSE_TYPE': 'DESCRIPTION_OF'}
        ]
        self.assertEqual(actual, expected)



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import unittest

from mock import MagicMock
from pyhocon import ConfigTree

from databuilder.publisher.base_publisher import NoopPublisher, Publisher


class TestPublisher(unittest.TestCase):

    def testCallback(self) -> None:
        publisher = NoopPublisher()
        callback = MagicMock()
        publisher.register_call_back(callback)
        publisher.publish()

        self.assertTrue(callback.on_success.called)

    def testFailureCallback(self) -> None:
        publisher = FailedPublisher()
        callback = MagicMock()
        publisher.register_call_back(callback)

        try:
            publisher.publish()
        except Exception:
            pass

        self.assertTrue(callback.on_failure.called)


class FailedPublisher(Publisher):
    def __init__(self) -> None:
        super(FailedPublisher, self).__init__()

    def init(self, conf: ConfigTree) -> None:
        pass

    def publish_impl(self) -> None:
        raise Exception('Bomb')


if __name__ == '__main__':
    unittest.main()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import base64
import json
import logging
import unittest

from mock import MagicMock, patch
from pyhocon import ConfigFactory

from databuilder.extractor.presto_view_metadata_extractor import PrestoViewMetadataExtractor
from databuilder.extractor.sql_alchemy_extractor import SQLAlchemyExtractor
from databuilder.models.table_metadata import ColumnMetadata, TableMetadata


class TestPrestoViewMetadataExtractor(unittest.TestCase):
    def setUp(self) -> None:
        logging.basicConfig(level=logging.INFO)

        config_dict = {
            f'extractor.sqlalchemy.{SQLAlchemyExtractor.CONN_STRING}': 'TEST_CONNECTION'
        }
        self.conf = ConfigFactory.from_dict(config_dict)

    def test_extraction_with_empty_result(self) -> None:
        """
        Test Extraction with empty result from query
        """
        with patch.object(SQLAlchemyExtractor, '_get_connection'):
            extractor = PrestoViewMetadataExtractor()
            extractor.init(self.conf)

            results = extractor.extract()
            self.assertEqual(results, None)

    def test_extraction_with_multiple_views(self) -> None:
        with patch.object(SQLAlchemyExtractor, '_get_connection') as mock_connection:
            connection = MagicMock()
            mock_connection.return_value = connection
            sql_execute = MagicMock()
            connection.execute = sql_execute

            columns1 = {'columns': [{'name': 'xyz', 'type': 'varchar'},
                                    {'name': 'xyy', 'type': 'double'},
                                    {'name': 'aaa', 'type': 'int'},
                                    {'name': 'ab', 'type': 'varchar'}]}

            columns2 = {'columns': [{'name': 'xyy', 'type': 'varchar'},
                                    {'name': 'ab', 'type': 'double'},
                                    {'name': 'aaa', 'type': 'int'},
                                    {'name': 'xyz', 'type': 'varchar'}]}

            sql_execute.return_value = [
                {'tbl_id': 2,
                 'schema': 'test_schema2',
                 'name': 'test_view2',
                 'tbl_type': 'virtual_view',
                 'view_original_text': base64.b64encode(json.dumps(columns2).encode()).decode("utf-8")},
                {'tbl_id': 1,
                 'schema': 'test_schema1',
                 'name': 'test_view1',
                 'tbl_type': 'virtual_view',
                 'view_original_text': base64.b64encode(json.dumps(columns1).encode()).decode("utf-8")},
            ]

            extractor = PrestoViewMetadataExtractor()
            extractor.init(self.conf)
            actual_first_view = extractor.extract()
            expected_first_view = TableMetadata('presto', 'gold', 'test_schema2', 'test_view2', None,
                                                [ColumnMetadata(u'xyy', None, u'varchar', 0),
                                                 ColumnMetadata(u'ab', None, u'double', 1),
                                                 ColumnMetadata(u'aaa', None, u'int', 2),
                                                 ColumnMetadata(u'xyz', None, u'varchar', 3)],
                                                True)
            self.assertEqual(expected_first_view.__repr__(), actual_first_view.__repr__())

            actual_second_view = extractor.extract()
            expected_second_view = TableMetadata('presto', 'gold', 'test_schema1', 'test_view1', None,
                                                 [ColumnMetadata(u'xyz', None, u'varchar', 0),
                                                  ColumnMetadata(u'xyy', None, u'double', 1),
                                                  ColumnMetadata(u'aaa', None, u'int', 2),
                                                  ColumnMetadata(u'ab', None, u'varchar', 3)],
                                                 True)
            self.assertEqual(expected_second_view.__repr__(), actual_second_view.__repr__())

            self.assertIsNone(extractor.extract())


if __name__ == '__main__':
    unittest.main()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

"""
This is a example script demonstrating how to load data into Neo4j and
Elasticsearch without using an Airflow DAG.

It contains several jobs:
- `run_csv_job`: runs a job that extracts table data from a CSV, loads (writes)
  this into a different local directory as a csv, then publishes this data to
  neo4j.
- `run_table_column_job`: does the same thing as `run_csv_job`, but with a csv
  containing column data.
- `create_last_updated_job`: creates a job that gets the current time, dumps it
  into a predefined model schema, and publishes this to neo4j.
- `create_es_publisher_sample_job`: creates a job that extracts data from neo4j
  and pubishes it into elasticsearch.

For other available extractors, please take a look at
https://github.com/amundsen-io/amundsendatabuilder#list-of-extractors
"""

import logging
import os
import sys
import uuid

from elasticsearch import Elasticsearch
from pyhocon import ConfigFactory
from sqlalchemy.ext.declarative import declarative_base

from databuilder.extractor.csv_extractor import (
    CsvExtractor, CsvTableBadgeExtractor, CsvTableColumnExtractor,
)
from databuilder.extractor.es_last_updated_extractor import EsLastUpdatedExtractor
from databuilder.extractor.neo4j_search_data_extractor import Neo4jSearchDataExtractor
from databuilder.job.job import DefaultJob
from databuilder.loader.file_system_elasticsearch_json_loader import FSElasticsearchJSONLoader
from databuilder.loader.file_system_neo4j_csv_loader import FsNeo4jCSVLoader
from databuilder.publisher.elasticsearch_constants import (
    DASHBOARD_ELASTICSEARCH_INDEX_MAPPING, USER_ELASTICSEARCH_INDEX_MAPPING,
)
from databuilder.publisher.elasticsearch_publisher import ElasticsearchPublisher
from databuilder.publisher.neo4j_csv_publisher import Neo4jCsvPublisher
from databuilder.task.task import DefaultTask
from databuilder.transformer.base_transformer import ChainedTransformer, NoopTransformer
from databuilder.transformer.dict_to_model import MODEL_CLASS, DictToModel
from databuilder.transformer.generic_transformer import (
    CALLBACK_FUNCTION, FIELD_NAME, GenericTransformer,
)

es_host = os.getenv('CREDENTIALS_ELASTICSEARCH_PROXY_HOST', 'localhost')
neo_host = os.getenv('CREDENTIALS_NEO4J_PROXY_HOST', 'localhost')

es_port = os.getenv('CREDENTIALS_ELASTICSEARCH_PROXY_PORT', 9200)
neo_port = os.getenv('CREDENTIALS_NEO4J_PROXY_PORT', 7687)
if len(sys.argv) > 1:
    es_host = sys.argv[1]
if len(sys.argv) > 2:
    neo_host = sys.argv[2]

es = Elasticsearch([
    {'host': es_host, 'port': es_port},
])

Base = declarative_base()

NEO4J_ENDPOINT = f'bolt://{neo_host}:{neo_port}'

neo4j_endpoint = NEO4J_ENDPOINT

neo4j_user = 'neo4j'
neo4j_password = 'test'

LOGGER = logging.getLogger(__name__)


def run_csv_job(file_loc, job_name, model):
    tmp_folder = f'/var/tmp/amundsen/{job_name}'
    node_files_folder = f'{tmp_folder}/nodes'
    relationship_files_folder = f'{tmp_folder}/relationships'

    csv_extractor = CsvExtractor()
    csv_loader = FsNeo4jCSVLoader()

    task = DefaultTask(extractor=csv_extractor,
                       loader=csv_loader,
                       transformer=NoopTransformer())

    job_config = ConfigFactory.from_dict({
        'extractor.csv.file_location': file_loc,
        'extractor.csv.model_class': model,
        'loader.filesystem_csv_neo4j.node_dir_path': node_files_folder,
        'loader.filesystem_csv_neo4j.relationship_dir_path': relationship_files_folder,
        'loader.filesystem_csv_neo4j.delete_created_directories': True,
        'publisher.neo4j.node_files_directory': node_files_folder,
        'publisher.neo4j.relation_files_directory': relationship_files_folder,
        'publisher.neo4j.neo4j_endpoint': neo4j_endpoint,
        'publisher.neo4j.neo4j_user': neo4j_user,
        'publisher.neo4j.neo4j_password': neo4j_password,
        'publisher.neo4j.neo4j_encrypted': False,
        'publisher.neo4j.job_publish_tag': 'unique_tag',  # should use unique tag here like {ds}
    })

    DefaultJob(conf=job_config,
               task=task,
               publisher=Neo4jCsvPublisher()).launch()


def run_table_badge_job(table_path, badge_path):
    tmp_folder = '/var/tmp/amundsen/table_badge'
    node_files_folder = f'{tmp_folder}/nodes'
    relationship_files_folder = f'{tmp_folder}/relationships'
    extractor = CsvTableBadgeExtractor()
    csv_loader = FsNeo4jCSVLoader()
    task = DefaultTask(extractor=extractor,
                       loader=csv_loader,
                       transformer=NoopTransformer())
    job_config = ConfigFactory.from_dict({
        'extractor.csvtablebadge.table_file_location': table_path,
        'extractor.csvtablebadge.badge_file_location': badge_path,
        'loader.filesystem_csv_neo4j.node_dir_path': node_files_folder,
        'loader.filesystem_csv_neo4j.relationship_dir_path': relationship_files_folder,
        'loader.filesystem_csv_neo4j.delete_created_directories': True,
        'publisher.neo4j.node_files_directory': node_files_folder,
        'publisher.neo4j.relation_files_directory': relationship_files_folder,
        'publisher.neo4j.neo4j_endpoint': neo4j_endpoint,
        'publisher.neo4j.neo4j_user': neo4j_user,
        'publisher.neo4j.neo4j_password': neo4j_password,
        'publisher.neo4j.neo4j_encrypted': False,
        'publisher.neo4j.job_publish_tag': 'unique_tag_b',  # should use unique tag here like {ds}
    })
    job = DefaultJob(conf=job_config,
                     task=task,
                     publisher=Neo4jCsvPublisher())
    job.launch()


def run_table_column_job(table_path, column_path):
    tmp_folder = '/var/tmp/amundsen/table_column'
    node_files_folder = f'{tmp_folder}/nodes'
    relationship_files_folder = f'{tmp_folder}/relationships'
    extractor = CsvTableColumnExtractor()
    csv_loader = FsNeo4jCSVLoader()
    task = DefaultTask(extractor,
                       loader=csv_loader,
                       transformer=NoopTransformer())
    job_config = ConfigFactory.from_dict({
        'extractor.csvtablecolumn.table_file_location': table_path,
        'extractor.csvtablecolumn.column_file_location': column_path,
        'loader.filesystem_csv_neo4j.node_dir_path': node_files_folder,
        'loader.filesystem_csv_neo4j.relationship_dir_path': relationship_files_folder,
        'loader.filesystem_csv_neo4j.delete_created_directories': True,
        'publisher.neo4j.node_files_directory': node_files_folder,
        'publisher.neo4j.relation_files_directory': relationship_files_folder,
        'publisher.neo4j.neo4j_endpoint': neo4j_endpoint,
        'publisher.neo4j.neo4j_user': neo4j_user,
        'publisher.neo4j.neo4j_password': neo4j_password,
        'publisher.neo4j.neo4j_encrypted': False,
        'publisher.neo4j.job_publish_tag': 'unique_tag',  # should use unique tag here like {ds}
    })
    job = DefaultJob(conf=job_config,
                     task=task,
                     publisher=Neo4jCsvPublisher())
    job.launch()


def create_last_updated_job():
    # loader saves data to these folders and publisher reads it from here
    tmp_folder = '/var/tmp/amundsen/last_updated_data'
    node_files_folder = f'{tmp_folder}/nodes'
    relationship_files_folder = f'{tmp_folder}/relationships'

    task = DefaultTask(extractor=EsLastUpdatedExtractor(),
                       loader=FsNeo4jCSVLoader())

    job_config = ConfigFactory.from_dict({
        'extractor.es_last_updated.model_class':
            'databuilder.models.es_last_updated.ESLastUpdated',

        'loader.filesystem_csv_neo4j.node_dir_path': node_files_folder,
        'loader.filesystem_csv_neo4j.relationship_dir_path': relationship_files_folder,
        'publisher.neo4j.node_files_directory': node_files_folder,
        'publisher.neo4j.relation_files_directory': relationship_files_folder,
        'publisher.neo4j.neo4j_endpoint': neo4j_endpoint,
        'publisher.neo4j.neo4j_user': neo4j_user,
        'publisher.neo4j.neo4j_password': neo4j_password,
        'publisher.neo4j.neo4j_encrypted': False,
        'publisher.neo4j.job_publish_tag': 'unique_lastupdated_tag',  # should use unique tag here like {ds}
    })

    return DefaultJob(conf=job_config,
                      task=task,
                      publisher=Neo4jCsvPublisher())


def _str_to_list(str_val):
    return str_val.split(',')


def create_dashboard_tables_job():
    # loader saves data to these folders and publisher reads it from here
    tmp_folder = '/var/tmp/amundsen/dashboard_table'
    node_files_folder = f'{tmp_folder}/nodes'
    relationship_files_folder = f'{tmp_folder}/relationships'

    csv_extractor = CsvExtractor()
    csv_loader = FsNeo4jCSVLoader()

    generic_transformer = GenericTransformer()
    dict_to_model_transformer = DictToModel()
    transformer = ChainedTransformer(transformers=[generic_transformer, dict_to_model_transformer],
                                     is_init_transformers=True)

    task = DefaultTask(extractor=csv_extractor,
                       loader=csv_loader,
                       transformer=transformer)
    publisher = Neo4jCsvPublisher()

    job_config = ConfigFactory.from_dict({
        f'{csv_extractor.get_scope()}.file_location': 'example/sample_data/sample_dashboard_table.csv',
        f'{transformer.get_scope()}.{generic_transformer.get_scope()}.{FIELD_NAME}': 'table_ids',
        f'{transformer.get_scope()}.{generic_transformer.get_scope()}.{CALLBACK_FUNCTION}': _str_to_list,
        f'{transformer.get_scope()}.{dict_to_model_transformer.get_scope()}.{MODEL_CLASS}':
            'databuilder.models.dashboard.dashboard_table.DashboardTable',
        f'{csv_loader.get_scope()}.node_dir_path': node_files_folder,
        f'{csv_loader.get_scope()}.relationship_dir_path': relationship_files_folder,
        f'{csv_loader.get_scope()}.delete_created_directories': True,
        f'{publisher.get_scope()}.node_files_directory': node_files_folder,
        f'{publisher.get_scope()}.relation_files_directory': relationship_files_folder,
        f'{publisher.get_scope()}.neo4j_endpoint': neo4j_endpoint,
        f'{publisher.get_scope()}.neo4j_user': neo4j_user,
        f'{publisher.get_scope()}.neo4j_password': neo4j_password,
        f'{publisher.get_scope()}.neo4j_encrypted': False,
        f'{publisher.get_scope()}.job_publish_tag': 'unique_tag',  # should use unique tag here like {ds}
    })

    return DefaultJob(conf=job_config,
                      task=task,
                      publisher=publisher)


def create_es_publisher_sample_job(elasticsearch_index_alias='table_search_index',
                                   elasticsearch_doc_type_key='table',
                                   model_name='databuilder.models.table_elasticsearch_document.TableESDocument',
                                   entity_type='table',
                                   elasticsearch_mapping=None):
    """
    :param elasticsearch_index_alias:  alias for Elasticsearch used in
                                       amundsensearchlibrary/search_service/config.py as an index
    :param elasticsearch_doc_type_key: name the ElasticSearch index is prepended with. Defaults to `table` resulting in
                                       `table_{uuid}`
    :param model_name:                 the Databuilder model class used in transporting between Extractor and Loader
    :param entity_type:                Entity type handed to the `Neo4jSearchDataExtractor` class, used to determine
                                       Cypher query to extract data from Neo4j. Defaults to `table`.
    :param elasticsearch_mapping:      Elasticsearch field mapping "DDL" handed to the `ElasticsearchPublisher` class,
                                       if None is given (default) it uses the `Table` query baked into the Publisher
    """
    # loader saves data to this location and publisher reads it from here
    extracted_search_data_path = '/var/tmp/amundsen/search_data.json'

    task = DefaultTask(loader=FSElasticsearchJSONLoader(),
                       extractor=Neo4jSearchDataExtractor(),
                       transformer=NoopTransformer())

    # elastic search client instance
    elasticsearch_client = es
    # unique name of new index in Elasticsearch
    elasticsearch_new_index_key = f'{elasticsearch_doc_type_key}_{uuid.uuid4()}'

    job_config = ConfigFactory.from_dict({
        'extractor.search_data.entity_type': entity_type,
        'extractor.search_data.extractor.neo4j.graph_url': neo4j_endpoint,
        'extractor.search_data.extractor.neo4j.model_class': model_name,
        'extractor.search_data.extractor.neo4j.neo4j_auth_user': neo4j_user,
        'extractor.search_data.extractor.neo4j.neo4j_auth_pw': neo4j_password,
        'extractor.search_data.extractor.neo4j.neo4j_encrypted': False,
        'loader.filesystem.elasticsearch.file_path': extracted_search_data_path,
        'loader.filesystem.elasticsearch.mode': 'w',
        'publisher.elasticsearch.file_path': extracted_search_data_path,
        'publisher.elasticsearch.mode': 'r',
        'publisher.elasticsearch.client': elasticsearch_client,
        'publisher.elasticsearch.new_index': elasticsearch_new_index_key,
        'publisher.elasticsearch.doc_type': elasticsearch_doc_type_key,
        'publisher.elasticsearch.alias': elasticsearch_index_alias,
    })

    # only optionally add these keys, so need to dynamically `put` them
    if elasticsearch_mapping:
        job_config.put(f'publisher.elasticsearch.{ElasticsearchPublisher.ELASTICSEARCH_MAPPING_CONFIG_KEY}',
                       elasticsearch_mapping)

    job = DefaultJob(conf=job_config,
                     task=task,
                     publisher=ElasticsearchPublisher())
    return job


if __name__ == "__main__":
    # Uncomment next line to get INFO level logging
    # logging.basicConfig(level=logging.INFO)

    run_table_column_job('example/sample_data/sample_table.csv', 'example/sample_data/sample_col.csv')
    run_table_badge_job('example/sample_data/sample_table.csv', 'example/sample_data/sample_badges.csv')
    run_csv_job('example/sample_data/sample_table_column_stats.csv', 'test_table_column_stats',
                'databuilder.models.table_stats.TableColumnStats')
    run_csv_job('example/sample_data/sample_table_programmatic_source.csv', 'test_programmatic_source',
                'databuilder.models.table_metadata.TableMetadata')
    run_csv_job('example/sample_data/sample_watermark.csv', 'test_watermark_metadata',
                'databuilder.models.watermark.Watermark')
    run_csv_job('example/sample_data/sample_table_owner.csv', 'test_table_owner_metadata',
                'databuilder.models.table_owner.TableOwner')
    run_csv_job('example/sample_data/sample_column_usage.csv', 'test_usage_metadata',
                'databuilder.models.column_usage_model.ColumnUsageModel')
    run_csv_job('example/sample_data/sample_user.csv', 'test_user_metadata',
                'databuilder.models.user.User')
    run_csv_job('example/sample_data/sample_application.csv', 'test_application_metadata',
                'databuilder.models.application.Application')
    run_csv_job('example/sample_data/sample_source.csv', 'test_source_metadata',
                'databuilder.models.table_source.TableSource')
    run_csv_job('example/sample_data/sample_tags.csv', 'test_tag_metadata',
                'databuilder.models.table_metadata.TagMetadata')
    run_csv_job('example/sample_data/sample_table_last_updated.csv', 'test_table_last_updated_metadata',
                'databuilder.models.table_last_updated.TableLastUpdated')
    run_csv_job('example/sample_data/sample_schema_description.csv', 'test_schema_description',
                'databuilder.models.schema.schema.SchemaModel')
    run_csv_job('example/sample_data/sample_dashboard_base.csv', 'test_dashboard_base',
                'databuilder.models.dashboard.dashboard_metadata.DashboardMetadata')
    run_csv_job('example/sample_data/sample_dashboard_usage.csv', 'test_dashboard_usage',
                'databuilder.models.dashboard.dashboard_usage.DashboardUsage')
    run_csv_job('example/sample_data/sample_dashboard_owner.csv', 'test_dashboard_owner',
                'databuilder.models.dashboard.dashboard_owner.DashboardOwner')
    run_csv_job('example/sample_data/sample_dashboard_query.csv', 'test_dashboard_query',
                'databuilder.models.dashboard.dashboard_query.DashboardQuery')
    run_csv_job('example/sample_data/sample_dashboard_last_execution.csv', 'test_dashboard_last_execution',
                'databuilder.models.dashboard.dashboard_execution.DashboardExecution')
    run_csv_job('example/sample_data/sample_dashboard_last_modified.csv', 'test_dashboard_last_modified',
                'databuilder.models.dashboard.dashboard_last_modified.DashboardLastModifiedTimestamp')

    create_dashboard_tables_job().launch()

    create_last_updated_job().launch()

    job_es_table = create_es_publisher_sample_job(
        elasticsearch_index_alias='table_search_index',
        elasticsearch_doc_type_key='table',
        entity_type='table',
        model_name='databuilder.models.table_elasticsearch_document.TableESDocument')
    job_es_table.launch()

    job_es_user = create_es_publisher_sample_job(
        elasticsearch_index_alias='user_search_index',
        elasticsearch_doc_type_key='user',
        model_name='databuilder.models.user_elasticsearch_document.UserESDocument',
        entity_type='user',
        elasticsearch_mapping=USER_ELASTICSEARCH_INDEX_MAPPING)
    job_es_user.launch()

    job_es_dashboard = create_es_publisher_sample_job(
        elasticsearch_index_alias='dashboard_search_index',
        elasticsearch_doc_type_key='dashboard',
        model_name='databuilder.models.dashboard_elasticsearch_document.DashboardESDocument',
        entity_type='dashboard',
        elasticsearch_mapping=DASHBOARD_ELASTICSEARCH_INDEX_MAPPING)
    job_es_dashboard.launch()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import unittest
from unittest.mock import ANY

from databuilder.models.graph_serializable import (
    NODE_KEY, NODE_LABEL, RELATION_END_KEY, RELATION_END_LABEL, RELATION_REVERSE_TYPE, RELATION_START_KEY,
    RELATION_START_LABEL, RELATION_TYPE,
)
from databuilder.models.table_stats import TableColumnStats
from databuilder.serializers import (
    mysql_serializer, neo4_serializer, neptune_serializer,
)
from databuilder.serializers.neptune_serializer import (
    METADATA_KEY_PROPERTY_NAME, NEPTUNE_CREATION_TYPE_JOB, NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME_BULK_LOADER_FORMAT,
    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT, NEPTUNE_HEADER_ID, NEPTUNE_HEADER_LABEL,
    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT, NEPTUNE_RELATIONSHIP_HEADER_FROM,
    NEPTUNE_RELATIONSHIP_HEADER_TO,
)


class TestTableStats(unittest.TestCase):

    def setUp(self) -> None:
        super(TestTableStats, self).setUp()
        self.table_stats = TableColumnStats(table_name='base.test',
                                            col_name='col',
                                            stat_name='avg',
                                            stat_val='1',
                                            start_epoch='1',
                                            end_epoch='2',)

        self.expected_node_results = [{
            NODE_KEY: 'hive://gold.base/test/col/avg/',
            NODE_LABEL: 'Stat',
            'stat_val': '1',
            'stat_type': 'avg',
            'start_epoch': '1',
            'end_epoch': '2',
        }]

        self.expected_relation_results = [{
            RELATION_START_KEY: 'hive://gold.base/test/col/avg/',
            RELATION_START_LABEL: 'Stat',
            RELATION_END_KEY: 'hive://gold.base/test/col',
            RELATION_END_LABEL: 'Column',
            RELATION_TYPE: 'STAT_OF',
            RELATION_REVERSE_TYPE: 'STAT'
        }]

    def test_get_table_stat_model_key(self) -> None:
        table_stats = self.table_stats.get_table_stat_model_key()
        self.assertEqual(table_stats, 'hive://gold.base/test/col/avg/')

    def test_get_col_key(self) -> None:
        metadata = self.table_stats.get_col_key()
        self.assertEqual(metadata, 'hive://gold.base/test/col')

    def test_create_nodes(self) -> None:
        actual = []
        node = self.table_stats.create_next_node()
        while node:
            serialized_node = neo4_serializer.serialize_node(node)
            actual.append(serialized_node)
            node = self.table_stats.create_next_node()

        self.assertEqual(actual, self.expected_node_results)

    def test_create_relation(self) -> None:
        actual = []
        relation = self.table_stats.create_next_relation()
        while relation:
            serialized_relation = neo4_serializer.serialize_relationship(relation)
            actual.append(serialized_relation)
            relation = self.table_stats.create_next_relation()

        self.assertEqual(actual, self.expected_relation_results)

    def test_create_nodes_neptune(self) -> None:
        actual = []
        next_node = self.table_stats.create_next_node()
        while next_node:
            serialized_node = neptune_serializer.convert_node(next_node)
            actual.append(serialized_node)
            next_node = self.table_stats.create_next_node()

        expected_neptune_nodes = [{
            NEPTUNE_HEADER_ID: 'Stat:hive://gold.base/test/col/avg/',
            METADATA_KEY_PROPERTY_NAME: 'Stat:hive://gold.base/test/col/avg/',
            NEPTUNE_HEADER_LABEL: 'Stat',
            NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
            NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB,
            'stat_val:String(single)': '1',
            'stat_type:String(single)': 'avg',
            'start_epoch:String(single)': '1',
            'end_epoch:String(single)': '2',
        }]

        self.assertEqual(actual, expected_neptune_nodes)

    def test_create_relation_neptune(self) -> None:
        self.expected_relation_result = {
            RELATION_START_KEY: 'hive://gold.base/test/col/avg/',
            RELATION_START_LABEL: 'Stat',
            RELATION_END_KEY: 'hive://gold.base/test/col',
            RELATION_END_LABEL: 'Column',
            RELATION_TYPE: 'STAT_OF',
            RELATION_REVERSE_TYPE: 'STAT'
        }

        expected = [
            [
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Stat:hive://gold.base/test/col/avg/',
                        to_vertex_id='Column:hive://gold.base/test/col',
                        label='STAT_OF'
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Stat:hive://gold.base/test/col/avg/',
                        to_vertex_id='Column:hive://gold.base/test/col',
                        label='STAT_OF'
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: 'Stat:hive://gold.base/test/col/avg/',
                    NEPTUNE_RELATIONSHIP_HEADER_TO: 'Column:hive://gold.base/test/col',
                    NEPTUNE_HEADER_LABEL: 'STAT_OF',
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                },
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Column:hive://gold.base/test/col',
                        to_vertex_id='Stat:hive://gold.base/test/col/avg/',
                        label='STAT'
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Column:hive://gold.base/test/col',
                        to_vertex_id='Stat:hive://gold.base/test/col/avg/',
                        label='STAT'
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: 'Column:hive://gold.base/test/col',
                    NEPTUNE_RELATIONSHIP_HEADER_TO: 'Stat:hive://gold.base/test/col/avg/',
                    NEPTUNE_HEADER_LABEL: 'STAT',
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                }
            ]
        ]

        actual = []
        next_relation = self.table_stats.create_next_relation()
        while next_relation:
            serialized_relation = neptune_serializer.convert_relationship(next_relation)
            actual.append(serialized_relation)
            next_relation = self.table_stats.create_next_relation()

        self.assertListEqual(actual, expected)

    def test_create_records(self) -> None:
        expected = [{
            'rk': 'hive://gold.base/test/col/avg/',
            'stat_val': '1',
            'stat_type': 'avg',
            'start_epoch': '1',
            'end_epoch': '2',
            'column_rk': 'hive://gold.base/test/col'
        }]

        actual = []
        record = self.table_stats.create_next_record()
        while record:
            serialized_record = mysql_serializer.serialize_record(record)
            actual.append(serialized_record)
            record = self.table_stats.create_next_record()

        self.assertEqual(actual, expected)



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

"""
This is a example script for extracting Delta Lake Metadata Results
"""

from pyhocon import ConfigFactory
from pyspark.sql import SparkSession

from databuilder.extractor.delta_lake_metadata_extractor import DeltaLakeMetadataExtractor
from databuilder.job.job import DefaultJob
from databuilder.loader.file_system_neo4j_csv_loader import FsNeo4jCSVLoader
from databuilder.models.table_metadata import DESCRIPTION_NODE_LABEL
from databuilder.publisher import neo4j_csv_publisher
from databuilder.publisher.neo4j_csv_publisher import Neo4jCsvPublisher
from databuilder.task.task import DefaultTask

# NEO4J cluster endpoints
NEO4J_ENDPOINT = 'bolt://localhost:7687/'

neo4j_endpoint = NEO4J_ENDPOINT

neo4j_user = 'neo4j'
neo4j_password = 'test'
cluster_key = 'my_delta_environment'
database = 'delta'
# Or set to empty to do all
schema_list = ['schema1', 'schema2']
# Or set to empty list if you don't have any schemas that you don't want to process
exclude_list = ['bad_schema']


def create_delta_lake_job_config():
    tmp_folder = '/var/tmp/amundsen/table_metadata'
    node_files_folder = f'{tmp_folder}/nodes/'
    relationship_files_folder = f'{tmp_folder}/relationships/'

    job_config = ConfigFactory.from_dict({
        f'extractor.delta_lake_table_metadata.{DeltaLakeMetadataExtractor.CLUSTER_KEY}': cluster_key,
        f'extractor.delta_lake_table_metadata.{DeltaLakeMetadataExtractor.DATABASE_KEY}': database,
        f'extractor.delta_lake_table_metadata.{DeltaLakeMetadataExtractor.SCHEMA_LIST_KEY}': schema_list,
        f'extractor.delta_lake_table_metadata.{DeltaLakeMetadataExtractor.EXCLUDE_LIST_SCHEMAS_KEY}': exclude_list,
        f'loader.filesystem_csv_neo4j.{FsNeo4jCSVLoader.NODE_DIR_PATH}': node_files_folder,
        f'loader.filesystem_csv_neo4j.{FsNeo4jCSVLoader.RELATION_DIR_PATH}': relationship_files_folder,
        f'publisher.neo4j.{neo4j_csv_publisher.NODE_FILES_DIR}': node_files_folder,
        f'publisher.neo4j.{neo4j_csv_publisher.RELATION_FILES_DIR}': relationship_files_folder,
        f'publisher.neo4j.{neo4j_csv_publisher.NEO4J_END_POINT_KEY}': neo4j_endpoint,
        f'publisher.neo4j.{neo4j_csv_publisher.NEO4J_USER}': neo4j_user,
        f'publisher.neo4j.{neo4j_csv_publisher.NEO4J_PASSWORD}': neo4j_password,
        f'publisher.neo4j.{neo4j_csv_publisher.NEO4J_CREATE_ONLY_NODES}': [DESCRIPTION_NODE_LABEL],
        f'publisher.neo4j.job_publish_tag': 'some_unique_tag'  # TO-DO unique tag must be added
    })
    return job_config


if __name__ == "__main__":
    # This assumes you are running on a spark cluster (for example databricks cluster)
    # that is configured with a hive metastore that
    # has pointers to all of your delta tables
    # Because of this, this code CANNOT run as a normal python operator on airflow.
    spark = SparkSession.builder.appName("Amundsen Delta Lake Metadata Extraction").getOrCreate()
    job_config = create_delta_lake_job_config()
    dExtractor = DeltaLakeMetadataExtractor()
    dExtractor.set_spark(spark)
    job = DefaultJob(conf=job_config,
                     task=DefaultTask(extractor=dExtractor, loader=FsNeo4jCSVLoader()),
                     publisher=Neo4jCsvPublisher())
    job.launch()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import unittest
from typing import (
    Iterable, Iterator, Union,
)

from amundsen_rds.models import RDSModel
from sqlalchemy import (
    Column, ForeignKey, String,
)
from sqlalchemy.ext.declarative import declarative_base

from databuilder.models.table_serializable import TableSerializable
from databuilder.serializers import mysql_serializer

Base = declarative_base()


class TestTableSerializable(unittest.TestCase):

    def test_table_serializable(self) -> None:
        actors = [Actor('Tom Cruise'), Actor('Meg Ryan')]
        movie = Movie('Top Gun', actors)

        actual = []
        node_row = movie.next_record()
        while node_row:
            actual.append(mysql_serializer.serialize_record(node_row))
            node_row = movie.next_record()

        expected = [
            {
                'rk': 'movie://Top Gun',
                'name': 'Top Gun'
            },
            {
                'rk': 'actor://Tom Cruise',
                'name': 'Tom Cruise'
            },
            {
                'movie_rk': 'movie://Top Gun',
                'actor_rk': 'actor://Tom Cruise'
            },
            {
                'rk': 'actor://Meg Ryan',
                'name': 'Meg Ryan'
            },
            {
                'movie_rk': 'movie://Top Gun',
                'actor_rk': 'actor://Meg Ryan'
            }
        ]

        self.assertEqual(expected, actual)


class RDSMovie(Base):  # type: ignore
    __tablename__ = 'movie'

    rk = Column(String(128), primary_key=True)
    name = Column(String(128))


class RDSActor(Base):  # type: ignore
    __tablename__ = 'actor'

    rk = Column(String(128), primary_key=True)
    name = Column(String(128))


class RDSMovieActor(Base):  # type: ignore
    __tablename__ = 'movie_actor'

    movie_rk = Column(String(128), ForeignKey('movie.rk'), primary_key=True)
    actor_rk = Column(String(128), ForeignKey('actor.rk'), primary_key=True)


class Actor(object):
    KEY_FORMAT = 'actor://{}'

    def __init__(self, name: str) -> None:
        self.name = name


class Movie(TableSerializable):
    KEY_FORMAT = 'movie://{}'

    def __init__(self,
                 name: str,
                 actors: Iterable[Actor]) -> None:
        self._name = name
        self._actors = actors
        self._record_iter = self._create_record_iterator()

    def create_next_record(self) -> Union[RDSModel, None]:
        try:
            return next(self._record_iter)
        except StopIteration:
            return None

    def _create_record_iterator(self) -> Iterator[RDSModel]:
        movie_record = RDSMovie(
            rk=Movie.KEY_FORMAT.format(self._name),
            name=self._name
        )
        yield movie_record

        for actor in self._actors:
            actor_record = RDSActor(
                rk=Actor.KEY_FORMAT.format(actor.name),
                name=actor.name
            )
            yield actor_record

            movie_actor_record = RDSMovieActor(
                movie_rk=Movie.KEY_FORMAT.format(self._name),
                actor_rk=Actor.KEY_FORMAT.format(actor.name)
            )
            yield movie_actor_record


if __name__ == '__main__':
    unittest.main()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import os
import unittest
from collections import OrderedDict
from csv import DictReader
from operator import itemgetter
from os import listdir
from os.path import isfile, join
from typing import (
    Any, Callable, Dict, Iterable,
)

from freezegun import freeze_time
from pyhocon import ConfigFactory

from databuilder.job.base_job import Job
from databuilder.loader.file_system_neptune_csv_loader import FSNeptuneCSVLoader
from tests.unit.models.test_graph_serializable import (
    Actor, City, Movie,
)


class FileSystemNeptuneCSVLoaderTest(unittest.TestCase):
    def setUp(self) -> None:
        prefix = '/var/tmp/TestFileSystemNeptuneCSVLoader'
        self._conf = ConfigFactory.from_dict(
            {
                FSNeptuneCSVLoader.NODE_DIR_PATH: '{}/{}'.format(prefix, 'nodes'),
                FSNeptuneCSVLoader.RELATION_DIR_PATH: '{}/{}'.format(prefix, 'relationships'),
                FSNeptuneCSVLoader.SHOULD_DELETE_CREATED_DIR: True,
                FSNeptuneCSVLoader.FORCE_CREATE_DIR: True,
                FSNeptuneCSVLoader.JOB_PUBLISHER_TAG: 'TESTED'
            }
        )

    def tearDown(self) -> None:
        Job.closer.close()

    @freeze_time("2020-09-01 01:01:00")
    def test_load(self) -> None:
        actors = [Actor('Tom Cruise'), Actor('Meg Ryan')]
        cities = [City('San Diego'), City('Oakland')]
        movie = Movie('Top Gun', actors, cities)

        loader = FSNeptuneCSVLoader()
        loader.init(self._conf)
        loader.load(movie)

        loader.close()

        expected_node_path = '{}/../resources/fs_neptune_csv_loader/nodes'.format(
            os.path.join(os.path.dirname(__file__))
        )
        expected_nodes = self._get_csv_rows(
            expected_node_path,
            itemgetter('~id')
        )
        actual_nodes = self._get_csv_rows(
            self._conf.get_string(FSNeptuneCSVLoader.NODE_DIR_PATH),
            itemgetter('~id')
        )
        self.maxDiff = None
        self.assertEqual(expected_nodes, actual_nodes)

        expected_rel_path = '{}/../resources/fs_neptune_csv_loader/relationships'.format(
            os.path.join(os.path.dirname(__file__))
        )
        expected_relations = self._get_csv_rows(
            expected_rel_path,
            itemgetter('~id')
        )
        actual_relations = self._get_csv_rows(
            self._conf.get_string(FSNeptuneCSVLoader.RELATION_DIR_PATH),
            itemgetter('~id')
        )
        self.assertListEqual(list(expected_relations), list(actual_relations))

    def _get_csv_rows(self, path: str, sorting_key_getter: Callable) -> Iterable[Dict[str, Any]]:
        files = [join(path, f) for f in listdir(path) if isfile(join(path, f))]

        result = []
        for f in files:
            with open(f, 'r') as f_input:
                reader = DictReader(f_input)
                for row in reader:
                    result.append(OrderedDict(sorted(row.items())))

        return sorted(result, key=sorting_key_getter)


if __name__ == '__main__':
    unittest.main()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
import unittest

from mock import MagicMock, patch
from pyhocon import ConfigFactory

from databuilder import Scoped
from databuilder.extractor.kafka_source_extractor import KafkaSourceExtractor


class TestKafkaSourceExtractor(unittest.TestCase):
    def setUp(self) -> None:
        logging.basicConfig(level=logging.INFO)
        config_dict = {
            f'extractor.kafka_source.consumer_config': {'"group.id"': 'consumer-group', '"enable.auto.commit"': False},
            f'extractor.kafka_source.{KafkaSourceExtractor.RAW_VALUE_TRANSFORMER}':
                'databuilder.transformer.base_transformer.NoopTransformer',
            f'extractor.kafka_source.{KafkaSourceExtractor.TOPIC_NAME_LIST}': ['test-topic'],
            f'extractor.kafka_source.{KafkaSourceExtractor.CONSUMER_TOTAL_TIMEOUT_SEC}': 1,
        }
        self.conf = ConfigFactory.from_dict(config_dict)

    def test_consume_success(self) -> None:
        kafka_extractor = KafkaSourceExtractor()
        kafka_extractor.init(Scoped.get_scoped_conf(conf=self.conf,
                                                    scope=kafka_extractor.get_scope()))

        with patch.object(kafka_extractor, 'consumer') as mock_consumer:
            mock_poll = MagicMock()
            mock_poll.error.return_value = False
            # only return once
            mock_poll.value.side_effect = ['msg']
            mock_consumer.poll.return_value = mock_poll

            records = kafka_extractor.consume()
            self.assertEqual(len(records), 1)

    def test_consume_fail(self) -> None:
        kafka_extractor = KafkaSourceExtractor()
        kafka_extractor.init(Scoped.get_scoped_conf(conf=self.conf,
                                                    scope=kafka_extractor.get_scope()))

        with patch.object(kafka_extractor, 'consumer') as mock_consumer:
            mock_poll = MagicMock()
            mock_poll.error.return_value = True
            mock_consumer.poll.return_value = mock_poll

            records = kafka_extractor.consume()
            self.assertEqual(len(records), 0)



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import unittest
from unittest.mock import ANY

from databuilder.models.graph_serializable import (
    NODE_KEY, NODE_LABEL, RELATION_END_KEY, RELATION_END_LABEL, RELATION_REVERSE_TYPE, RELATION_START_KEY,
    RELATION_START_LABEL, RELATION_TYPE,
)
from databuilder.models.table_last_updated import TableLastUpdated
from databuilder.models.timestamp import timestamp_constants
from databuilder.serializers import (
    mysql_serializer, neo4_serializer, neptune_serializer,
)
from databuilder.serializers.neptune_serializer import (
    METADATA_KEY_PROPERTY_NAME, NEPTUNE_CREATION_TYPE_JOB, NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME_BULK_LOADER_FORMAT,
    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT, NEPTUNE_HEADER_ID, NEPTUNE_HEADER_LABEL,
    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT, NEPTUNE_RELATIONSHIP_HEADER_FROM,
    NEPTUNE_RELATIONSHIP_HEADER_TO,
)


class TestTableLastUpdated(unittest.TestCase):

    def setUp(self) -> None:
        super(TestTableLastUpdated, self).setUp()

        self.tableLastUpdated = TableLastUpdated(table_name='test_table',
                                                 last_updated_time_epoch=25195665,
                                                 schema='default')

        self.expected_node_results = [{
            NODE_KEY: 'hive://gold.default/test_table/timestamp',
            NODE_LABEL: 'Timestamp',
            'last_updated_timestamp:UNQUOTED': 25195665,
            timestamp_constants.TIMESTAMP_PROPERTY + ":UNQUOTED": 25195665,
            'name': 'last_updated_timestamp'
        }]

        self.expected_relation_results = [{
            RELATION_START_KEY: 'hive://gold.default/test_table',
            RELATION_START_LABEL: 'Table',
            RELATION_END_KEY: 'hive://gold.default/test_table/timestamp',
            RELATION_END_LABEL: 'Timestamp',
            RELATION_TYPE: 'LAST_UPDATED_AT',
            RELATION_REVERSE_TYPE: 'LAST_UPDATED_TIME_OF'
        }]

    def test_get_table_model_key(self) -> None:
        table = self.tableLastUpdated.get_table_model_key()
        self.assertEqual(table, 'hive://gold.default/test_table')

    def test_get_last_updated_model_key(self) -> None:
        last_updated = self.tableLastUpdated.get_last_updated_model_key()
        self.assertEqual(last_updated, 'hive://gold.default/test_table/timestamp')

    def test_create_nodes(self) -> None:
        actual = []
        node = self.tableLastUpdated.create_next_node()
        while node:
            serialize_node = neo4_serializer.serialize_node(node)
            actual.append(serialize_node)
            node = self.tableLastUpdated.create_next_node()

        self.assertEqual(actual, self.expected_node_results)

    def test_create_nodes_neptune(self) -> None:
        node_id = TableLastUpdated.LAST_UPDATED_NODE_LABEL + ":" + self.tableLastUpdated.get_last_updated_model_key()
        expected_nodes = [{
            NEPTUNE_HEADER_ID: node_id,
            METADATA_KEY_PROPERTY_NAME: node_id,
            NEPTUNE_HEADER_LABEL: TableLastUpdated.LAST_UPDATED_NODE_LABEL,
            NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
            NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB,
            'name:String(single)': 'last_updated_timestamp',
            'last_updated_timestamp:Long(single)': 25195665,
            timestamp_constants.TIMESTAMP_PROPERTY + ":Long(single)": 25195665,
        }]

        actual = []
        next_node = self.tableLastUpdated.create_next_node()
        while next_node:
            next_node_serialized = neptune_serializer.convert_node(next_node)
            actual.append(next_node_serialized)
            next_node = self.tableLastUpdated.create_next_node()

        self.assertEqual(actual, expected_nodes)

    def test_create_relation(self) -> None:
        actual = []
        relation = self.tableLastUpdated.create_next_relation()
        while relation:
            serialized_relation = neo4_serializer.serialize_relationship(relation)
            actual.append(serialized_relation)
            relation = self.tableLastUpdated.create_next_relation()

    def test_create_relation_neptune(self) -> None:
        actual = []
        next_relation = self.tableLastUpdated.create_next_relation()
        while next_relation:
            next_relation_serialized = neptune_serializer.convert_relationship(next_relation)
            actual.append(next_relation_serialized)
            next_relation = self.tableLastUpdated.create_next_relation()

        expected = [
            [
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Table:hive://gold.default/test_table',
                        to_vertex_id='Timestamp:hive://gold.default/test_table/timestamp',
                        label='LAST_UPDATED_AT'
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Table:hive://gold.default/test_table',
                        to_vertex_id='Timestamp:hive://gold.default/test_table/timestamp',
                        label='LAST_UPDATED_AT'
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: 'Table:hive://gold.default/test_table',
                    NEPTUNE_RELATIONSHIP_HEADER_TO: 'Timestamp:hive://gold.default/test_table/timestamp',
                    NEPTUNE_HEADER_LABEL: 'LAST_UPDATED_AT',
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                },
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Timestamp:hive://gold.default/test_table/timestamp',
                        to_vertex_id='Table:hive://gold.default/test_table',
                        label='LAST_UPDATED_TIME_OF'
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Timestamp:hive://gold.default/test_table/timestamp',
                        to_vertex_id='Table:hive://gold.default/test_table',
                        label='LAST_UPDATED_TIME_OF'
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: 'Timestamp:hive://gold.default/test_table/timestamp',
                    NEPTUNE_RELATIONSHIP_HEADER_TO: 'Table:hive://gold.default/test_table',
                    NEPTUNE_HEADER_LABEL: 'LAST_UPDATED_TIME_OF',
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                }
            ]
        ]

        self.assertEqual(actual, expected)

    def test_create_records(self) -> None:
        expected = [{
            'rk': 'hive://gold.default/test_table/timestamp',
            'last_updated_timestamp': 25195665,
            'timestamp': 25195665,
            'name': 'last_updated_timestamp',
            'table_rk': 'hive://gold.default/test_table'
        }]

        actual = []
        record = self.tableLastUpdated.create_next_record()
        while record:
            serialized_record = mysql_serializer.serialize_record(record)
            actual.append(serialized_record)
            record = self.tableLastUpdated.create_next_record()

        self.assertEqual(expected, actual)



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
import tempfile
import unittest
from typing import Dict

from pyhocon import ConfigFactory
# patch whole class to avoid actually calling for boto3.client during tests
from pyspark.sql import SparkSession
from pyspark.sql.catalog import Table

from databuilder import Scoped
from databuilder.extractor.delta_lake_metadata_extractor import (
    DeltaLakeMetadataExtractor, ScrapedColumnMetadata, ScrapedTableMetadata,
)
from databuilder.models.table_metadata import ColumnMetadata, TableMetadata


class TestDeltaLakeExtractor(unittest.TestCase):

    def setUp(self) -> None:
        logging.basicConfig(level=logging.INFO)
        self.spark = SparkSession.builder \
            .appName("Amundsen Delta Lake Metadata Extraction") \
            .master("local") \
            .config("spark.jars.packages", "io.delta:delta-core_2.12:0.7.0") \
            .config("spark.sql.warehouse.dir", tempfile.TemporaryDirectory()) \
            .config("spark.sql.extensions", "io.delta.sql.DeltaSparkSessionExtension") \
            .config("spark.sql.catalog.spark_catalog", "org.apache.spark.sql.delta.catalog.DeltaCatalog") \
            .getOrCreate()
        self.config_dict = {
            f'extractor.delta_lake_table_metadata.{DeltaLakeMetadataExtractor.CLUSTER_KEY}': 'test_cluster',
            f'extractor.delta_lake_table_metadata.{DeltaLakeMetadataExtractor.SCHEMA_LIST_KEY}': [],
            f'extractor.delta_lake_table_metadata.{DeltaLakeMetadataExtractor.EXCLUDE_LIST_SCHEMAS_KEY}': [],
            f'extractor.delta_lake_table_metadata.{DeltaLakeMetadataExtractor.DATABASE_KEY}': 'test_database',
            f'extractor.delta_lake_table_metadata.{DeltaLakeMetadataExtractor.DELTA_TABLES_ONLY}': False
        }
        conf = ConfigFactory.from_dict(self.config_dict)
        self.dExtractor = DeltaLakeMetadataExtractor()
        self.dExtractor.init(Scoped.get_scoped_conf(conf=conf, scope=self.dExtractor.get_scope()))
        self.dExtractor.set_spark(self.spark)
        self.setUpSchemas()

    def setUpSchemas(self) -> None:
        self.spark.sql("create schema if not exists test_schema1")
        self.spark.sql("create schema if not exists test_schema2")
        self.spark.sql("create table if not exists test_schema1.test_table1 (a string, b int) using delta")
        self.spark.sql("create table if not exists "
                       "test_schema1.test_table3 (c boolean, d float) using delta partitioned by (c)")
        self.spark.sql("create table if not exists test_schema2.test_parquet (a string) using parquet")
        self.spark.sql("create table if not exists test_schema2.test_table2 (a2 string, b2 double) using delta")
        # TODO do we even need to support views and none delta tables in this case?
        self.spark.sql("create view if not exists test_schema2.test_view1 as (select * from test_schema2.test_table2)")

    def test_get_all_schemas(self) -> None:
        '''Tests getting all schemas'''
        actual_schemas = self.dExtractor.get_schemas([])
        self.assertEqual(["default", "test_schema1", "test_schema2"], actual_schemas)

    def test_get_all_schemas_with_exclude(self) -> None:
        '''Tests the exclude list'''
        actual_schemas = self.dExtractor.get_schemas(["default"])
        self.assertEqual(["test_schema1", "test_schema2"], actual_schemas)

    def test_get_all_tables(self) -> None:
        '''Tests table fetching'''
        actual = [x.name for x in self.dExtractor.get_all_tables(["test_schema1", "default"])]
        self.assertEqual(["test_table1", "test_table3"], actual)

    def test_scrape_table_detail(self) -> None:
        '''Test Table Detail Scraping'''
        actual = self.dExtractor.scrape_table_detail("test_schema1.test_table1")
        expected: Dict = {'createdAt': None,
                          'description': None,
                          'format': 'delta',
                          'id': None,
                          'lastModified': None,
                          'location': None,
                          'minReaderVersion': 1,
                          'minWriterVersion': 2,
                          'name': 'test_schema1.test_table1',
                          'numFiles': 0,
                          'partitionColumns': [],
                          'properties': {},
                          'sizeInBytes': 0}
        self.assertIsNotNone(actual)
        if actual:
            self.assertEqual(actual.keys(), expected.keys())
            self.assertEqual(actual['name'], expected['name'])
            self.assertEqual(actual['format'], expected['format'])

    def test_scrape_view_detail(self) -> None:
        actual = self.dExtractor.scrape_view_detail("test_schema2.test_view1")
        self.assertIsNotNone(actual)
        expected = {'Created By': 'Spark 3.0.1',
                    'Created Time': None,
                    'Database': 'test_schema2',
                    'Last Access': 'UNKNOWN',
                    'Table': 'test_view1',
                    'Table Properties': '[view.catalogAndNamespace.numParts=2, '
                                        'view.query.out.col.0=a2, view.query.out.numCols=2, '
                                        'view.query.out.col.1=b2, '
                                        'view.catalogAndNamespace.part.0=spark_catalog, '
                                        'view.catalogAndNamespace.part.1=default]',
                    'Type': 'VIEW',
                    'View Catalog and Namespace': 'spark_catalog.default',
                    'View Original Text': '(select * from test_schema2.test_table2)',
                    'View Query Output Columns': '[a2, b2]',
                    'View Text': '(select * from test_schema2.test_table2)'}
        if actual:
            actual['Created Time'] = None
            self.assertEqual(actual, expected)

    def test_fetch_partitioned_delta_columns(self) -> None:
        actual = self.dExtractor.fetch_columns("test_schema1", "test_table3")
        partition_column = ScrapedColumnMetadata(name="c", description=None, data_type="boolean", sort_order=0)
        partition_column.set_is_partition(True)
        expected = [partition_column,
                    ScrapedColumnMetadata(name="d", description=None, data_type="float", sort_order=1)]
        for a, b in zip(actual, expected):
            self.assertEqual(a, b)

    def test_fetch_delta_columns(self) -> None:
        actual = self.dExtractor.fetch_columns("test_schema1", "test_table1")
        expected = [ScrapedColumnMetadata(name="a", description=None, data_type="string", sort_order=0),
                    ScrapedColumnMetadata(name="b", description=None, data_type="int", sort_order=1)]
        for a, b in zip(actual, expected):
            self.assertEqual(a, b)

    def test_scrape_tables(self) -> None:
        table = Table(name="test_table1", database="test_schema1", description=None,
                      tableType="delta", isTemporary=False)
        actual = self.dExtractor.scrape_table(table)

        expected = ScrapedTableMetadata(schema="test_schema1", table="test_table1")
        expected.set_columns([ScrapedColumnMetadata(name="a", description=None, data_type="string", sort_order=0),
                              ScrapedColumnMetadata(name="b", description=None, data_type="int", sort_order=1)])
        if actual is not None:
            self.assertEqual(expected.schema, actual.schema)
            self.assertEqual(expected.table, actual.table)
            self.assertEqual(expected.columns, actual.columns)
            self.assertEqual(expected.failed_to_scrape, actual.failed_to_scrape)
            self.assertEqual(expected.is_view, actual.is_view)
            self.assertIsNotNone(actual.table_detail)
        else:
            self.assertIsNotNone(actual)

    def test_create_table_metadata(self) -> None:
        scraped = ScrapedTableMetadata(schema="test_schema1", table="test_table1")
        scraped.set_columns([ScrapedColumnMetadata(name="a", description=None, data_type="string", sort_order=0),
                             ScrapedColumnMetadata(name="b", description=None, data_type="int", sort_order=1)])
        created_metadata = self.dExtractor.create_table_metadata(scraped)
        expected = TableMetadata("test_database", "test_cluster", "test_schema1", "test_table1", description=None,
                                 columns=[ColumnMetadata("a", None, "string", 0),
                                          ColumnMetadata("b", None, "int", 1)])
        self.assertEqual(str(expected), str(created_metadata))

    def test_create_last_updated(self) -> None:
        scraped_table = self.dExtractor.scrape_table(Table("test_table1", "test_schema1", None, "delta", False))
        actual_last_updated = None
        if scraped_table:
            actual_last_updated = self.dExtractor.create_table_last_updated(scraped_table)
        self.assertIsNotNone(actual_last_updated)

    def test_extract(self) -> None:
        ret = []
        data = self.dExtractor.extract()
        while data is not None:
            ret.append(data)
            data = self.dExtractor.extract()
        self.assertEqual(len(ret), 8)

    def test_extract_with_only_specific_schemas(self) -> None:
        self.config_dict = {
            f'extractor.delta_lake_table_metadata.{DeltaLakeMetadataExtractor.CLUSTER_KEY}': 'test_cluster',
            f'extractor.delta_lake_table_metadata.{DeltaLakeMetadataExtractor.SCHEMA_LIST_KEY}': ['test_schema2'],
            f'extractor.delta_lake_table_metadata.{DeltaLakeMetadataExtractor.EXCLUDE_LIST_SCHEMAS_KEY}': [],
            f'extractor.delta_lake_table_metadata.{DeltaLakeMetadataExtractor.DATABASE_KEY}': 'test_database'
        }
        conf = ConfigFactory.from_dict(self.config_dict)
        self.dExtractor.init(Scoped.get_scoped_conf(conf=conf,
                                                    scope=self.dExtractor.get_scope()))
        ret = []
        data = self.dExtractor.extract()
        while data is not None:
            ret.append(data)
            data = self.dExtractor.extract()
        self.assertEqual(len(ret), 2)

    def test_extract_when_excluding(self) -> None:
        self.config_dict = {
            f'extractor.delta_lake_table_metadata.{DeltaLakeMetadataExtractor.CLUSTER_KEY}': 'test_cluster',
            f'extractor.delta_lake_table_metadata.{DeltaLakeMetadataExtractor.SCHEMA_LIST_KEY}': [],
            f'extractor.delta_lake_table_metadata.{DeltaLakeMetadataExtractor.EXCLUDE_LIST_SCHEMAS_KEY}':
                ['test_schema2'],
            f'extractor.delta_lake_table_metadata.{DeltaLakeMetadataExtractor.DATABASE_KEY}': 'test_database'
        }
        conf = ConfigFactory.from_dict(self.config_dict)
        self.dExtractor.init(Scoped.get_scoped_conf(conf=conf,
                                                    scope=self.dExtractor.get_scope()))
        ret = []
        data = self.dExtractor.extract()
        while data is not None:
            ret.append(data)
            data = self.dExtractor.extract()
        self.assertEqual(len(ret), 4)

    def test_table_does_not_exist(self) -> None:
        table = Table(name="test_table5", database="test_schema1", description=None,
                      tableType="delta", isTemporary=False)
        actual = self.dExtractor.scrape_table(table)
        self.assertIsNone(actual)

    def test_scrape_all_tables(self) -> None:
        tables = [Table(name="test_table1", database="test_schema1", description=None,
                        tableType="delta", isTemporary=False),
                  Table(name="test_table3", database="test_schema1", description=None,
                        tableType="delta", isTemporary=False)]
        actual = self.dExtractor.scrape_all_tables(tables)
        self.assertEqual(2, len(actual))


if __name__ == '__main__':
    unittest.main()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import unittest

from pyhocon import ConfigFactory

from databuilder.transformer.template_variable_substitution_transformer import (
    FIELD_NAME, TEMPLATE, TemplateVariableSubstitutionTransformer,
)


class TestTemplateVariableSubstitutionTransformer(unittest.TestCase):

    def test_conversion(self) -> None:

        transformer = TemplateVariableSubstitutionTransformer()
        config = ConfigFactory.from_dict({
            FIELD_NAME: 'baz',
            TEMPLATE: 'Hello {foo}'
        })
        transformer.init(conf=config)

        actual = transformer.transform({'foo': 'bar'})
        expected = {
            'foo': 'bar',
            'baz': 'Hello bar'
        }
        self.assertDictEqual(expected, actual)


if __name__ == '__main__':
    unittest.main()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
import random
import unittest
from typing import (
    Any, Dict, List,
)

from mock import patch

from databuilder.extractor.dashboard.redash.redash_dashboard_utils import (
    RedashPaginatedRestApiQuery, generate_dashboard_description, get_auth_headers, get_text_widgets,
    get_visualization_widgets, sort_widgets,
)
from databuilder.rest_api.base_rest_api_query import EmptyRestApiQuerySeed

logging.basicConfig(level=logging.INFO)


class TestRedashDashboardUtils(unittest.TestCase):
    def test_sort_widgets(self) -> None:
        widgets = [
            {
                'text': 'a',
                'options': {}
            },
            {
                'text': 'b',
                'options': {'position': {'row': 1, 'col': 1}}
            },
            {
                'text': 'c',
                'options': {'position': {'row': 1, 'col': 2}}
            },
            {
                'text': 'd',
                'options': {'position': {'row': 2, 'col': 1}}
            }
        ]
        random.shuffle(widgets)
        sorted_widgets = sort_widgets(widgets)
        self.assertListEqual([widget['text'] for widget in sorted_widgets], ['a', 'b', 'c', 'd'])

    def test_widget_filters(self) -> None:
        widgets: List[Dict[str, Any]] = [
            {'text': 'asdf', 'options': {'ex': 1}},
            {'text': 'asdf', 'options': {'ex': 2}},
            {'visualization': {}, 'options': {'ex': 1}},
            {'visualization': {}, 'options': {'ex': 2}},
            {'visualization': {}, 'options': {'ex': 3}}
        ]
        self.assertEqual(len(get_text_widgets(widgets)), 2)
        self.assertEqual(len(get_visualization_widgets(widgets)), 3)

    def test_text_widget_props(self) -> None:
        widget_data = {
            'text': 'asdf'
        }
        widget = get_text_widgets([widget_data])[0]
        self.assertEqual(widget.text, 'asdf')

    def test_visualization_widget_props(self) -> None:
        widget_data = {
            'visualization': {
                'query': {
                    'id': 123,
                    'data_source_id': 1,
                    'query': 'SELECT 2+2 FROM DUAL',
                    'name': 'Test'
                },
                'id': 12345,
                'name': 'test_widget',
                'type': 'CHART'
            }
        }
        widget = get_visualization_widgets([widget_data])[0]

        self.assertEqual(widget.query_id, 123)
        self.assertEqual(widget.data_source_id, 1)
        self.assertEqual(widget.raw_query, 'SELECT 2+2 FROM DUAL')
        self.assertEqual(widget.query_name, 'Test')
        self.assertEqual(widget.visualization_id, 12345)
        self.assertEqual(widget.visualization_name, 'test_widget')
        self.assertEqual(widget.visualization_type, 'CHART')

    def test_descriptions_from_text(self) -> None:
        text_widgets = get_text_widgets([
            {'text': 'T1'},
            {'text': 'T2'}
        ])
        viz_widgets = get_visualization_widgets([
            {
                'visualization': {
                    'query': {
                        'id': 1,
                        'data_source_id': 1,
                        'name': 'Q1',
                        'query': 'n/a'
                    }
                }
            },
            {
                'visualization': {
                    'query': {
                        'id': 2,
                        'data_source_id': 1,
                        'name': 'Q2',
                        'query': 'n/a'
                    }
                }
            }
        ])

        # both text and viz widgets
        desc1 = generate_dashboard_description(text_widgets, viz_widgets)
        self.assertTrue('T1' in desc1)
        self.assertTrue('T2' in desc1)
        self.assertTrue('Q1' not in desc1)

        # only text widgets
        desc2 = generate_dashboard_description(text_widgets, [])
        self.assertEqual(desc1, desc2)

        # only viz widgets
        desc3 = generate_dashboard_description([], viz_widgets)
        self.assertTrue('Q1' in desc3)
        self.assertTrue('Q2' in desc3)

        # no widgets
        desc4 = generate_dashboard_description([], [])
        self.assertTrue('empty' in desc4)

    def test_descriptions_remove_duplicate(self) -> None:
        viz_widgets = get_visualization_widgets([
            {
                'visualization': {
                    'query': {
                        'id': 1,
                        'data_source_id': 1,
                        'name': 'same_query_name',
                        'query': 'n/a'
                    }
                }
            },
            {
                'visualization': {
                    'query': {
                        'id': 2,
                        'data_source_id': 1,
                        'name': 'same_query_name',
                        'query': 'n/a'
                    }
                }
            }
        ])
        desc1 = generate_dashboard_description([], viz_widgets)
        self.assertEqual('A dashboard containing the following queries:\n\n- same_query_name', desc1)

    def test_auth_headers(self) -> None:
        headers = get_auth_headers('testkey')
        self.assertTrue('testkey' in headers['Authorization'])

    def test_paginated_rest_api_query(self) -> None:
        paged_content = [
            {
                'page': 1,
                'page_size': 5,
                'count': 12,
                'results': [{'test': True}] * 5
            },
            {
                'page': 2,
                'page_size': 5,
                'count': 12,
                'results': [{'test': True}] * 5
            },
            {
                'page': 3,
                'page_size': 5,
                'count': 12,
                'results': [{'test': True}] * 2
            },
            {
                'page': 4,
                'page_size': 5,
                'count': 12,
                'results': []
            }
        ]

        with patch('databuilder.rest_api.rest_api_query.requests.get') as mock_get:
            # .json() is called twice (ugh), so we have to double each page
            mock_get.return_value.json.side_effect = [page for page in paged_content for page in [page] * 2]

            q = RedashPaginatedRestApiQuery(query_to_join=EmptyRestApiQuerySeed(),
                                            url='example.com',
                                            json_path='results[*].[test]',
                                            params={},
                                            field_names=['test'],
                                            skip_no_result=True)
            n_records = 0
            for record in q.execute():
                self.assertEqual(record['test'], True)
                n_records += 1

            self.assertEqual(n_records, 12)


if __name__ == '__main__':
    unittest.main()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import unittest
from unittest.mock import ANY

from databuilder.models.graph_serializable import (
    RELATION_END_KEY, RELATION_END_LABEL, RELATION_REVERSE_TYPE, RELATION_START_KEY, RELATION_START_LABEL,
    RELATION_TYPE,
)
from databuilder.models.table_source import TableSource
from databuilder.serializers import (
    mysql_serializer, neo4_serializer, neptune_serializer,
)
from databuilder.serializers.neptune_serializer import (
    METADATA_KEY_PROPERTY_NAME, NEPTUNE_CREATION_TYPE_JOB,
    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT, NEPTUNE_HEADER_ID, NEPTUNE_HEADER_LABEL,
    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT, NEPTUNE_RELATIONSHIP_HEADER_FROM,
    NEPTUNE_RELATIONSHIP_HEADER_TO,
)

DB = 'hive'
SCHEMA = 'base'
TABLE = 'test'
CLUSTER = 'default'
SOURCE = '/etl/sql/file.py'


class TestTableSource(unittest.TestCase):

    def setUp(self) -> None:
        super(TestTableSource, self).setUp()
        self.table_source = TableSource(db_name='hive',
                                        schema=SCHEMA,
                                        table_name=TABLE,
                                        cluster=CLUSTER,
                                        source=SOURCE)

        self.start_key = f'{DB}://{CLUSTER}.{SCHEMA}/{TABLE}/_source'
        self.end_key = f'{DB}://{CLUSTER}.{SCHEMA}/{TABLE}'

    def test_get_source_model_key(self) -> None:
        source = self.table_source.get_source_model_key()
        self.assertEqual(source, f'{DB}://{CLUSTER}.{SCHEMA}/{TABLE}/_source')

    def test_get_metadata_model_key(self) -> None:
        metadata = self.table_source.get_metadata_model_key()
        self.assertEqual(metadata, 'hive://default.base/test')

    def test_create_nodes(self) -> None:
        expected_nodes = [{
            'LABEL': 'Source',
            'KEY': f'{DB}://{CLUSTER}.{SCHEMA}/{TABLE}/_source',
            'source': SOURCE,
            'source_type': 'github'
        }]

        actual = []
        node = self.table_source.create_next_node()
        while node:
            serialized_node = neo4_serializer.serialize_node(node)
            actual.append(serialized_node)
            node = self.table_source.create_next_node()

        self.assertEqual(expected_nodes, actual)

    def test_create_relation(self) -> None:
        expected_relations = [{
            RELATION_START_KEY: self.start_key,
            RELATION_START_LABEL: TableSource.LABEL,
            RELATION_END_KEY: self.end_key,
            RELATION_END_LABEL: 'Table',
            RELATION_TYPE: TableSource.SOURCE_TABLE_RELATION_TYPE,
            RELATION_REVERSE_TYPE: TableSource.TABLE_SOURCE_RELATION_TYPE
        }]

        actual = []
        relation = self.table_source.create_next_relation()
        while relation:
            serialized_relation = neo4_serializer.serialize_relationship(relation)
            actual.append(serialized_relation)
            relation = self.table_source.create_next_relation()

        self.assertEqual(expected_relations, actual)

    def test_create_relation_neptune(self) -> None:
        actual = []
        relation = self.table_source.create_next_relation()
        while relation:
            serialized_relation = neptune_serializer.convert_relationship(relation)
            actual.append(serialized_relation)
            relation = self.table_source.create_next_relation()

        expected = [
            [
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id="Source:" + self.start_key,
                        to_vertex_id="Table:" + self.end_key,
                        label=TableSource.SOURCE_TABLE_RELATION_TYPE
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id="Source:" + self.start_key,
                        to_vertex_id="Table:" + self.end_key,
                        label=TableSource.SOURCE_TABLE_RELATION_TYPE
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: "Source:" + self.start_key,
                    NEPTUNE_RELATIONSHIP_HEADER_TO: "Table:" + self.end_key,
                    NEPTUNE_HEADER_LABEL: TableSource.SOURCE_TABLE_RELATION_TYPE,
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                },
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id="Table:" + self.end_key,
                        to_vertex_id="Source:" + self.start_key,
                        label=TableSource.TABLE_SOURCE_RELATION_TYPE
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id="Table:" + self.end_key,
                        to_vertex_id="Source:" + self.start_key,
                        label=TableSource.TABLE_SOURCE_RELATION_TYPE
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: "Table:" + self.end_key,
                    NEPTUNE_RELATIONSHIP_HEADER_TO: "Source:" + self.start_key,
                    NEPTUNE_HEADER_LABEL: TableSource.TABLE_SOURCE_RELATION_TYPE,
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                }
            ]
        ]

        self.assertListEqual(expected, actual)

    def test_create_records(self) -> None:
        expected = [{
            'rk': self.table_source.get_source_model_key(),
            'source': self.table_source.source,
            'source_type': self.table_source.source_type,
            'table_rk': self.table_source.get_metadata_model_key()
        }]

        actual = []
        record = self.table_source.create_next_record()
        while record:
            serialized_record = mysql_serializer.serialize_record(record)
            actual.append(serialized_record)
            record = self.table_source.create_next_record()

        self.assertEqual(expected, actual)
